# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Least Cost Xmission Command Line Interface
"""
import click
import logging
import os
import json

import numpy as np
import geopandas as gpd

from rex.utilities.loggers import init_logger, init_mult, create_dirs
from rex.utilities.cli_dtypes import STR, INT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties

from reVX.config.least_cost_xmission import LeastCostPathsConfig
from reVX.least_cost_xmission.config import XmissionConfig
from reVX.least_cost_xmission.least_cost_paths import (LeastCostPaths,
                                                       ReinforcementPaths)
from reVX.least_cost_xmission.least_cost_xmission import ba_mapper
from reVX.least_cost_xmission.config import TRANS_LINE_CAT, SUBSTATION_CAT
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='LeastCostPaths', type=STR,
              show_default=True,
              help='Job name.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Least Cost Paths Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['NAME'] = name


@main.command()
def valid_config_keys():
    """
    Echo the valid Least Cost Paths config keys
    """
    click.echo(', '.join(get_class_properties(LeastCostPathsConfig)))


def run_local(ctx, config):
    """
    Run Least Cost Paths locally using config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : reVX.config.least_cost_xmission.LeastCostPathsConfig
        Least Cost Paths config object.
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               cost_fpath=config.cost_fpath,
               features_fpath=config.features_fpath,
               network_nodes_fpath=config.network_nodes_fpath,
               transmission_lines_fpath=config.transmission_lines_fpath,
               capacity_class=config.capacity_class,
               xmission_config=config.xmission_config,
               start_index=0, step_index=1,
               barrier_mult=config.barrier_mult,
               max_workers=config.execution_control.max_workers,
               save_paths=config.save_paths,
               out_dir=config.dirout,
               log_dir=config.log_directory,
               verbose=config.log_level)


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to AssemblyAreas config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run Least Cost Paths from a config.
    """

    config = LeastCostPathsConfig(config)

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if config.execution_control.option == 'local':
        run_local(ctx, config)
        return

    if config.execution_control.option != 'eagle':
        click.echo('Option "{}" is not supported'
                   .format(config.execution_control.option))
        return

    # No need to add index to file name
    if config.execution_control.nodes == 1:
        eagle(config)
        return

    name = config.name
    num_nodes = config.execution_control.nodes
    logger.info('Splitting features over {} SLURM jobs'.format(num_nodes))
    n_zfill = len(str(num_nodes))
    for i in range(num_nodes):
        config.name = '{}_{}'.format(name, str(i).zfill(n_zfill))
        eagle(config, start_index=i)


@main.command()
@click.option('--cost_fpath', '-cost', type=click.Path(exists=True),
              required=True,
              help=("Path to h5 file with cost rasters and other required "
                    "layers"))
@click.option('--features_fpath', '-feats', required=True,
              type=click.Path(exists=True),
              help="Path to GeoPackage with transmission features")
@click.option('--capacity_class', '-cap', type=str, required=True,
              help=("Capacity class of transmission features to connect "
                    "supply curve points to"))
@click.option('--network_nodes_fpath', '-nn', type=STR, show_default=True,
              default=None,
              help=("Path to Network Nodes GeoPackage. If given alongside "
                    "`transmission_lines_fpath`, reinforcement path cost "
                    "calculation is run."))
@click.option('--transmission_lines_fpath', '-tl', type=STR, show_default=True,
              default=None,
              help=("Path to Transmission lines GeoPackage. This file can "
                    "contain other features, but transmission lines must "
                    "be identified by {!r}. If given alongside "
                    "`network_nodes_fpath`, reinforcement path cost "
                    "calculation is run.".format(TRANS_LINE_CAT)))
@click.option('--xmission_config', '-xcfg', type=STR, show_default=True,
              default=None,
              help=("Path to transmission config .json"))
@click.option('--start_index', '-start', type=int,
              show_default=True, default=0,
              help=("Start index of features to run."))
@click.option('--step_index', '-step', type=int,
              show_default=True, default=1,
              help=("Step index of features to run."))
@click.option('--barrier_mult', '-bmult', type=float,
              show_default=True, default=100,
              help=("Transmission barrier multiplier, used when computing the "
                    "least cost tie-line path"))
@click.option('--max_workers', '-mw', type=INT,
              show_default=True, default=None,
              help=("Number of workers to use for processing, if 1 run in "
                    "serial, if None use all available cores"))
@click.option('--save_paths', '-paths', is_flag=True,
              help="Flag to save least cost path as a multi-line geometry")
@click.option('--out_dir', '-o', type=STR, default='./',
              show_default=True,
              help='Directory to save least cost Paths values to.')
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def local(ctx, cost_fpath, features_fpath, capacity_class, network_nodes_fpath,
          transmission_lines_fpath, xmission_config, start_index, step_index,
          barrier_mult, max_workers, save_paths, out_dir, log_dir, verbose):
    """
    Run Least Cost Paths on local hardware
    """
    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    create_dirs(out_dir)
    logger.info('Computing Least Cost Paths connections and writing them to {}'
                .format(out_dir))
    xmission_config = XmissionConfig(config=xmission_config)
    is_reinforcement_run = (network_nodes_fpath is not None
                            and transmission_lines_fpath is not None)
    if is_reinforcement_run:
        features = gpd.read_file(network_nodes_fpath)
        features, *__ = LeastCostPaths._map_to_costs(cost_fpath, features)
        indices = features.index[start_index::step_index]
        least_costs = ReinforcementPaths.run(cost_fpath, features_fpath,
                                             network_nodes_fpath,
                                             transmission_lines_fpath,
                                             capacity_class,
                                             xmission_config=xmission_config,
                                             barrier_mult=barrier_mult,
                                             indices=indices,
                                             save_paths=save_paths)
    else:
        features = gpd.read_file(features_fpath)
        features, *__ = LeastCostPaths._map_to_costs(cost_fpath, features)
        indices = features.index[start_index::step_index]
        least_costs = LeastCostPaths.run(cost_fpath, features_fpath,
                                         capacity_class,
                                         xmission_config=xmission_config,
                                         barrier_mult=barrier_mult,
                                         indices=indices,
                                         max_workers=max_workers,
                                         save_paths=save_paths)

    capacity_class = xmission_config._parse_cap_class(capacity_class)
    cap = xmission_config['power_classes'][capacity_class]
    kv = xmission_config.capacity_to_kv(capacity_class)
    fn_out = '{}_{}MW_{}kV'.format(name, cap, kv)
    fpath_out = os.path.join(out_dir, fn_out)
    if save_paths:
        fpath_out += '.gpkg'
        least_costs.to_file(fpath_out, driver="GPKG", index=False)
    else:
        fpath_out += '.csv'
        least_costs.to_csv(fpath_out, index=False)


@main.command()
@click.option('--features_fpath', '-feats', required=True,
              type=click.Path(exists=True),
              help="Path to GeoPackage with substation and transmission "
                   "features")
@click.option('--balancing_areas_fpath', '-ba', required=True,
              type=click.Path(exists=True),
              help=("Path to Balancing areas GeoPackage."))
@click.option('--out_file', '-of', default=None, type=STR,
              help='Name for output GeoPackage file.')
@click.pass_context
def map_ba(ctx, features_fpath, balancing_areas_fpath, out_file):
    """
    Map substation locations to balancing regions.

    This method also removes substations that do not meet the min 69 kV
    voltage requirement and adds {'min_volts', 'max_volts'} fields to
    the remaining substations.

    **IMPORTANT** This method DOES NOT clip the substations to your
    balancing area boundary. All substations will be mapped to their
    closest BA. It is your responsibility to remove any substations
    outside of the analysis region before calling this method.

    Doing the pre-processing step avoids any issues with substations
    being left out or double counted if we just clipped them to the
    balancing area shapes.
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)

    features = gpd.read_file(features_fpath)
    ba = gpd.read_file(balancing_areas_fpath).to_crs(features.crs)
    substations = (features[features.category == SUBSTATION_CAT]
                   .reset_index(drop=True).dropna(axis="columns", how="all"))

    logger.info("Mapping {:,d} substation locations to {:,d} balancing "
                "areas".format(substations.shape[0], ba.shape[0]))

    substations["ba_str"] = substations.centroid.apply(ba_mapper(ba))

    logger.info("Calculating min/max voltage for each substation...")
    bad_subs = np.zeros(len(substations), dtype=bool)
    for idx, row in substations.iterrows():
        lines = row['trans_gids']
        if isinstance(lines, str):
            lines = json.loads(lines)

        lines_mask = features['gid'].isin(lines)
        voltage = features.loc[lines_mask, 'voltage'].values

        if np.max(voltage) >= 69:
            substations.loc[idx, 'min_volts'] = np.min(voltage)
            substations.loc[idx, 'max_volts'] = np.max(voltage)
        else:
            bad_subs[idx] = True

    if any(bad_subs):
        msg = ("The following sub-stations do not have the minimum "
               "required voltage of 69 kV and will be dropped:\n{}"
               .format(substations.loc[bad_subs, 'gid']))
        logger.warning(msg)
        substations = substations.loc[~bad_subs].reset_index(drop=True)

    logger.info("Writing substation output to {!r}".format(out_file))
    substations.to_file(out_file, driver="GPKG", index=False)


def get_node_cmd(config, start_index=0):
    """
    Get the node CLI call for Least Cost Paths

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostPathsConfig
        Least Cost Paths config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """

    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-cost {}'.format(SLURM.s(config.cost_fpath)),
            '-feats {}'.format(SLURM.s(config.features_fpath)),
            '-nn {}'.format(SLURM.s(config.network_nodes_fpath)),
            '-tl {}'.format(SLURM.s(config.transmission_lines_fpath)),
            '-cap {}'.format(SLURM.s(config.capacity_class)),
            '-start {}'.format(SLURM.s(start_index)),
            '-step {}'.format(SLURM.s(config.execution_control.nodes or 1)),
            '-bmult {}'.format(SLURM.s(config.barrier_mult)),
            '-mw {}'.format(SLURM.s(config.execution_control.max_workers)),
            '-o {}'.format(SLURM.s(config.dirout)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            ]

    if config.save_paths:
        args.append('-paths')

    if config.log_level == logging.DEBUG:
        args.append('-v')

    cmd = ('python -m reVX.least_cost_xmission.least_cost_paths_cli {}'
           .format(' '.join(args)))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def eagle(config, start_index=0):
    """
    Run Least Cost Paths on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostPathsConfig
        Least Cost Paths config object.
    """

    cmd = get_node_cmd(config, start_index)
    name = config.name
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')

    slurm_manager = SLURM()

    logger.info('Running Least Cost Paths on Eagle with '
                'node name "{}"'.format(name))
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off Least Cost Paths "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off Least Cost Paths '
               '"{}". Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running Least Cost Paths CLI')
        raise



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
import os
import logging
import fiona
import geopandas as gpd

from reVX.setbacks.base import AbstractBaseSetbacks
from reVX.utilities.utilities import STATES_ABBR_MAP

logger = logging.getLogger(__name__)


class RoadSetbacks(AbstractBaseSetbacks):
    """
    Road setbacks
    """

    def parse_features(self):
        """
        Load roads from gdb file, convert to exclusions coordinate
        system.

        Warnings
        --------
        Use caution when calling this method, especially in multiple
        processes, as the returned feature files may be quite large.
        Reading 100 GB feature files in each of 36 sub-processes will
        quickly overwhelm your RAM.

        Returns
        -------
        roads : `geopandas.GeoDataFrame.sindex`
            Geometries for roads in gdb file, in exclusion coordinate
            system
        """
        lyr = fiona.listlayers(self._features)[0]
        roads = gpd.read_file(self._features, driver='FileGDB', layer=lyr)
        return roads.to_crs(crs=self._rasterizer.profile["crs"])

    @staticmethod
    def get_feature_paths(features_fpath):
        """
        Find all roads gdb files in roads_dir

        Parameters
        ----------
        features_fpath : str
            Path to state here streets gdb file or directory containing
            states gdb files. Used to identify roads to build setbacks
            from. Files should be by state.

        Returns
        -------
        file_paths : list
            List of file paths to all roads .gdp files in roads_dir
        """
        is_file = (features_fpath.endswith('.gdb')
                   or features_fpath.endswith('.gpkg'))
        if is_file:
            file_paths = [features_fpath]
        else:
            file_paths = []
            for file in sorted(os.listdir(features_fpath)):
                is_file = file.endswith('.gdb') or file.endswith('.gpkg')
                if is_file and file.startswith('Streets_USA'):
                    file_paths.append(os.path.join(features_fpath, file))

        return file_paths

    def _regulation_table_mask(self):
        """Return the regulation table mask for setback feature. """
        state = self._features.split('.')[0].split('_')[-1]
        if 'Abbr' not in self.regulations_table:
            states = self.regulations_table['State'].str.title()
            self.regulations_table['Abbr'] = states.map(STATES_ABBR_MAP)
        states = self.regulations_table['Abbr'] == state

        feature_types = {'roads', 'highways', 'highways 111'}
        features = self.regulations_table['Feature Type'].isin(feature_types)
        return states & features



# -*- coding: utf-8 -*-
"""
Setbacks CLI
"""
import click
from copy import deepcopy
import logging
import os
from pathlib import Path

from rex.utilities.loggers import init_mult, init_logger
from rex.utilities.cli_dtypes import STR, FLOAT, INT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties, dict_str_load

from reVX.config.setbacks import SetbacksConfig
from reVX.setbacks import SETBACKS
from reVX.setbacks.regulations import (validate_setback_regulations_input,
                                       select_setback_regulations)
from reVX.setbacks.setbacks_converter import parse_setbacks
from reVX.handlers.geotiff import Geotiff
from reVX.utilities import ExclusionsConverter
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default=SetbacksConfig.NAME,
              type=STR, show_default=True,
              help='Job name. Default is {!r}.'.format(SetbacksConfig.NAME))
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Setbacks Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid Setbacks config keys
    """
    click.echo(', '.join(get_class_properties(SetbacksConfig)))


@main.command()
@click.option('--tiff_dir', '-td', required=True,
              type=click.Path(exists=True),
              help=("Path to directory containing geotiffs to be merged."))
@click.option('--out_file', '-o', required=True,
              type=click.Path(),
              help=("Path to output tiff file."))
@click.option('--are_partial_inclusions', '-inclusions', is_flag=True,
              help=('Flag to indicate that the data in the tiff layers '
                    'represent partial inclusion values (i.e. 0.25 = 25% '
                    'included), NOT typical exclusion values (i.e. '
                    '1 = exclude pixel)'))
@click.pass_context
def merge(ctx, tiff_dir, out_file, are_partial_inclusions):
    """
    Combine setbacks geotiffs into a single exclusion (or inclusion) layer.

    This command assumes the data in separate files is non-overlapping.
    In other words, a file containing setbacks exclusions for Illinois
    should not contain any exclusions for Indiana, assuming the setbacks
    for Indiana are in a separate tif file in the same directory.
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)

    logger.info("Merging tiff files in {!r}".format(tiff_dir))
    setbacks = [path.as_posix() for path in Path(tiff_dir).glob("*.tif*")]
    if not setbacks:
        msg = ("Did not find any files ending in '.tif' in directory: {}"
               .format(tiff_dir))
        logger.error(msg)
        raise FileNotFoundError(msg)

    with Geotiff(setbacks[0]) as tif:
        profile = tif.profile

    out_file = Path(out_file).resolve().as_posix()
    combined_setbacks = parse_setbacks(
        setbacks, is_inclusion_layer=are_partial_inclusions)

    logger.info("Writing data to {!r}".format(out_file))
    ExclusionsConverter.write_geotiff(out_file, profile, combined_setbacks)


@main.command()
@click.option('--excl_fpath', '-excl', required=True,
              type=click.Path(exists=True),
              help=('Path to .h5 file containing exclusion layers, will also '
                    'be the location of any new setback layers'))
@click.option('--feature_type', '-ft', required=True, type=STR,
              help=('Setback feature type. Must be one of {}'
                    .format(set(SETBACKS.keys()))))
@click.option('--features_path', '-feats', required=True,
              type=click.Path(exists=True),
              help=('Path to:/n'
                    '- State level structure .geojson or directory containing '
                    'geotiff files./n'
                    '- State level roads .gdb or directory containing .gdb '
                    'files./n'
                    '-transmission or railroad CONUS wide .shp file./n'
                    '- State level parcel .gpkg file.'))
@click.option('--out_dir', '-o', required=True, type=STR,
              help=('Directory to save setbacks geotiff(s) into'))
@click.option('--hub_height', '-hh', default=None, type=FLOAT,
              help=('Turbine hub height(m), used along with rotor diameter to '
                    'compute blade tip height which is used to determine '
                    'setback distance. Must be provided if '
                    '`base_setback_dist` is not given'))
@click.option('--rotor_diameter', '-rd', default=None, type=FLOAT,
              help=('Turbine rotor diameter(m), used along with hub height to '
                    'compute blade tip height which is used to determine '
                    'setback distance. Must be provided if '
                    '`base_setback_dist` is not given'))
@click.option('--base_setback_dist', '-bsd', default=None, type=FLOAT,
              help=('Base setback distance value. Must be provided if '
                    'hub_height and rotor_diameter are not given'))
@click.option('--regs_fpath', '-regs', default=None, type=STR,
              show_default=True,
              help=('Path to regulations .csv file, if None create '
                    'generic setbacks using max - tip height * "multiplier", '
                    'by default None'))
@click.option('--multiplier', '-mult', default=None, type=FLOAT,
              show_default=True,
              help=('setback multiplier to use if regulations are not '
                    'supplied, if str, must a key in '
                    '{"high": 3, "moderate": 1.1}, if supplied along with '
                    'regs_fpath, will be ignored, multiplied with max-tip '
                    'height, by default None'))
@click.option('--weights_calculation_upscale_factor', '-wcuf',
              default=None, type=INT, show_default=True,
              help=('Scale factor to use for inclusion weights calculation. '
                    'See the `AbstractBaseSetbacks` documentation for more '
                    'details. By default None.'))
@click.option('--max_workers', '-mw', default=None, type=INT,
              show_default=True,
              help=('Number of workers to use for setback computation, if 1 '
                    'run in serial, if > 1 run in parallel with that many '
                    'workers, if None run in parallel on all available cores, '
                    'by default None'))
@click.option('--replace', '-r', is_flag=True,
              help=('Flag to replace local layer data with arr if layer '
                    'already exists in the exclusion .h5 file'))
@click.option('--hsds', '-hsds', is_flag=True,
              help=('Flag to use h5pyd to handle .h5 domain hosted on AWS '
                    'behind HSDS'))
@click.option('--out_layers', '-ol', type=STR, default=None,
              show_default=True,
              help=('String representation of a dictionary mapping feature '
                    'file names (with extension) to names of layers under '
                    'which exclusions should be saved in the "excl_fpath" '
                    '.h5 file. If "None" or empty dictionary, no layers are '
                    'saved to the h5 file.'))
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files. Default is out_dir.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def local(ctx, excl_fpath, feature_type, features_path, out_dir, hub_height,
          rotor_diameter, base_setback_dist, regs_fpath, multiplier,
          weights_calculation_upscale_factor, max_workers, replace, hsds,
          out_layers, log_dir, verbose):
    """
    Compute Setbacks locally
    """
    if feature_type not in SETBACKS:
        msg = ("feature_type must be one of: {}; got {!r}"
               .format(set(SETBACKS.keys()), feature_type))
        raise RuntimeError(msg)

    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(ctx.obj['NAME'], log_dir, modules=log_modules, verbose=verbose)

    # same check as the config in case someone invokes this from the
    # direct command line instead of a config file for some bizarre reason
    validate_setback_regulations_input(base_setback_dist, hub_height,
                                       rotor_diameter)

    logger.info('Computing setbacks from structures in {}'
                .format(features_path))
    logger.debug('Setbacks to be computed with:\n'
                 '- base_setback_dist = {}\n'
                 '- hub_height = {}\n'
                 '- rotor_diameter = {}\n'
                 '- regs_fpath = {}\n'
                 '- multiplier = {}\n'
                 '- using max_workers = {}\n'
                 '- replace layer if needed = {}\n'
                 '- weights calculation upscale factor = {}\n'
                 '- out_layers = {}\n'
                 .format(base_setback_dist, hub_height, rotor_diameter,
                         regs_fpath, multiplier, max_workers, replace,
                         weights_calculation_upscale_factor, out_layers))

    regulations = select_setback_regulations(base_setback_dist, hub_height,
                                             rotor_diameter, regs_fpath,
                                             multiplier)

    setbacks_class = SETBACKS[feature_type]
    wcuf = weights_calculation_upscale_factor
    if isinstance(out_layers, str):
        out_layers = dict_str_load(out_layers)

    setbacks_class.run(excl_fpath, features_path, out_dir, regulations,
                       weights_calculation_upscale_factor=wcuf,
                       max_workers=max_workers, replace=replace, hsds=hsds,
                       out_layers=out_layers)
    logger.info('Setbacks computed and written to {}'.format(out_dir))


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to Setbacks config json file.')
@click.pass_context
def from_config(ctx, config):
    """
    Run Setbacks from a config.
    """
    config = SetbacksConfig(config)

    if config.execution_control.option == 'local':
        run_local(ctx, config)

    if config.execution_control.option == 'eagle':
        eagle(config)


def run_local(ctx, config):
    """
    Run Setbacks locally from config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : `reVX.config.setbacks.SetbacksConfig`
        Setbacks config object.
    """
    ctx.obj['NAME'] = config.name
    wcuf = config.weights_calculation_upscale_factor
    ctx.invoke(local,
               excl_fpath=config.excl_fpath,
               feature_type=config.feature_type,
               features_path=config.features_path,
               out_dir=config.dirout,
               hub_height=config.hub_height,
               rotor_diameter=config.rotor_diameter,
               base_setback_dist=config.base_setback_dist,
               regs_fpath=config.regs_fpath,
               multiplier=config.multiplier,
               weights_calculation_upscale_factor=wcuf,
               max_workers=config.execution_control.max_workers,
               replace=config.replace,
               hsds=config.hsds,
               out_layers=config.out_layers,
               verbose=config.log_level==logging.DEBUG)


def eagle(config):
    """
    Run Setbacks on Eagle HPC.

    Parameters
    ----------
    config : `reVX.config.setbacks.SetbacksConfig`
        Setbacks config object.
    """
    features_path = config.features_path
    cls = SETBACKS[config.feature_type]
    features = cls.get_feature_paths(features_path)
    if not features:
        msg = ('No valid feature files were found at {}!'
               .format(features_path))
        logger.error(msg)
        raise FileNotFoundError(msg)

    for fpath in features:
        fpath_config = deepcopy(config)
        fpath_config['features_path'] = fpath
        launch_job(fpath_config)


def launch_job(config):
    """
    Launch job from config on SLURM

    Parameters
    ----------
    config : `reVX.config.setbacks.SetbacksConfig`
        Setbacks config object.
    """
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')
    name = os.path.basename(config.features_path).split('.')[0]
    name = "{}-{}".format(config.name, name)
    cmd = get_node_cmd(name, config)

    logger.info('Computing Setbacks on Eagle with '
                'node name "{}"'.format(name))
    slurm_manager = SLURM()
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off Setbacks job "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off Setbacks job "{}". '
               'Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


def get_node_cmd(name, config):
    """
    Get the node CLI call for the Setbacks computation

    Parameters
    ----------
    config : `reVX.config.setbacks.SetbacksConfig`
        Setbacks config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """
    wcuf = config.weights_calculation_upscale_factor
    args = ['-n {}'.format(SLURM.s(name)),
            'local',
            '-excl {}'.format(SLURM.s(config.excl_fpath)),
            '-ft {}'.format(SLURM.s(config.feature_type)),
            '-feats {}'.format(SLURM.s(config.features_path)),
            '-o {}'.format(SLURM.s(config.dirout)),
            '-regs {}'.format(SLURM.s(config.regs_fpath)),
            '-mult {}'.format(SLURM.s(config.multiplier)),
            '-wcuf {}'.format(SLURM.s(wcuf)),
            '-ol {}'.format(SLURM.s(config.out_layers)),
            '-mw {}'.format(SLURM.s(config.execution_control.max_workers)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            ]

    if config.base_setback_dist is None:
        args.append('-hh {}'.format(SLURM.s(config.hub_height)))
        args.append('-rd {}'.format(SLURM.s(config.rotor_diameter)))
    else:
        args.append('-bsd {}'.format(SLURM.s(config.base_setback_dist)))

    if config.replace:
        args.append('-r')

    if config.hsds:
        args.append('-hsds')

    if config.log_level == logging.DEBUG:
        args.append('-v')

    cmd = 'python -m reVX.setbacks.setbacks_cli {}'.format(' '.join(args))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running Setbacks CLI')
        raise



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
import logging
import os
import numpy as np
import geopandas as gpd

from rex.utilities import log_mem

from reVX.setbacks.base import AbstractBaseSetbacks


logger = logging.getLogger(__name__)


class ParcelSetbacks(AbstractBaseSetbacks):
    """Parcel setbacks - facilitates the use of negative buffers. """

    @staticmethod
    def compute_local_exclusions(regulation_value, cnty, *args):
        """Compute local features setbacks.

        This method will compute the setbacks using a county-specific
        regulations file that specifies either a static setback or a
        multiplier value that will be used along with the base setback
        distance to compute the setback.

        Parameters
        ----------
        regulation_value : float | int
            Setback distance in meters.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.
        features : geopandas.GeoDataFrame
            Features for the local county.
        feature_filter : callable
            A callable function that takes `features` and `cnty` as
            inputs and outputs a geopandas.GeoDataFrame with features
            clipped and/or localized to the input county.
        rasterizer : Rasterizer
            Instance of `Rasterizer` class used to rasterize the
            buffered county features.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        features, feature_filter, rasterizer = args
        logger.debug('- Computing setbacks for county FIPS {}'
                     .format(cnty.iloc[0]['FIPS']))
        log_mem(logger)
        features = feature_filter(features, cnty)
        negative_buffer = features.buffer(-1 * regulation_value)
        setbacks = features.buffer(0).difference(negative_buffer)
        return rasterizer.rasterize(list(setbacks))

    def compute_generic_exclusions(self, **__):
        """Compute generic setbacks.

        This method will compute the setbacks using a generic setback
        of `base_setback_dist * multiplier`.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        logger.info("Computing generic setbacks")
        if np.isclose(self._regulations.generic, 0):
            return self._rasterizer.rasterize(shapes=None)

        negative_buffer = self.features.buffer(-1 * self._regulations.generic)
        setbacks = self.features.buffer(0).difference(negative_buffer)
        return self._rasterizer.rasterize(list(setbacks))

    def _regulation_table_mask(self):
        """Return the regulation table mask for setback feature. """
        state = os.path.basename(self._features).split('.')[0]
        state = _get_state_name(state)
        states = self.regulations_table.State.apply(_get_state_name)
        states = states == state
        property_line = (self.regulations_table['Feature Type']
                         == 'property line')
        return states & property_line

    def parse_features(self):
        """Parse in parcel features.

        Warnings
        --------
        Use caution when calling this method, especially in multiple
        processes, as the returned feature files may be quite large.
        Reading 100 GB feature files in each of 36 sub-processes will
        quickly overwhelm your RAM.

        Returns
        -------
        `geopandas.GeoDataFrame`
            Geometries of features to setback from in exclusion
            coordinate system.
        """
        features = gpd.read_file(self._features)
        if features.crs is None:
            features = features.set_crs("EPSG:4326")
        return features.to_crs(crs=self._rasterizer.profile["crs"])


def _get_state_name(state):
    """Filter out non-alpha chars and casefold name"""
    return ''.join(filter(str.isalpha, state.casefold()))



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
import os
import re
import logging

from reVX.setbacks.base import AbstractBaseSetbacks

logger = logging.getLogger(__name__)


class StructureSetbacks(AbstractBaseSetbacks):
    """
    Structure setbacks
    """

    @staticmethod
    def _split_state_name(state_name):
        """
        Split state name at capitals to map .geojson files to
        regulations state names.

        Parameters
        ----------
        state_name : str
            State name from geojson files paths with out spaces

        Returns
        -------
        str
            State names with spaces added between Capitals (names) to
            match wind regulations state names.
        """
        state_name = ' '.join(a for a
                              in re.split(r'([A-Z][a-z]*)', state_name)
                              if a)

        return state_name

    @staticmethod
    def get_feature_paths(features_fpath):
        """
        Find all structures .geojson files in structures dir

        Parameters
        ----------
        features_fpath : str
            Path to structures geojson for a single state, or directory
            containing geojsons for all states. Used to identify
            structures to build setbacks from. Files should be by state.

        Returns
        -------
        file_paths : list
            List of file paths to all structures .geojson files in
            structures_dir
        """
        is_file = (features_fpath.endswith('.geojson')
                   or features_fpath.endswith('.gpkg'))
        if is_file:
            file_paths = [features_fpath]
        else:
            file_paths = []
            for file in sorted(os.listdir(features_fpath)):
                is_file = file.endswith('.geojson') or file.endswith('.gpkg')
                if is_file:
                    file_paths.append(os.path.join(features_fpath, file))

        return file_paths

    def _regulation_table_mask(self):
        """Return the regulation table mask for setback feature. """
        state_name = os.path.basename(self._features).split('.')[0]
        state = self._split_state_name(state_name)
        states = self.regulations_table["State"] == state
        structures = self.regulations_table['Feature Type'] == 'structures'
        not_ocb = (self.regulations_table['Comment']
                   != 'Occupied Community Buildings')
        return states & structures & not_ocb



# -*- coding: utf-8 -*-
"""
reVX Setbacks Configuration
"""
import logging

from reV.config.base_analysis_config import AnalysisConfig
from reVX.setbacks.regulations import validate_setback_regulations_input
from reVX.setbacks import SETBACKS

logger = logging.getLogger(__name__)


class SetbacksConfig(AnalysisConfig):
    """Config framework for setbacks"""

    NAME = 'Setbacks'
    REQUIREMENTS = ('excl_fpath', 'features_path', 'feature_type')

    def _preflight(self):
        """
        Run a preflight check for extra requirements based on feature type.
        """
        super()._preflight()
        validate_setback_regulations_input(
            base_setback_dist=self.base_setback_dist,
            hub_height=self.hub_height,
            rotor_diameter=self.rotor_diameter)

    @property
    def feature_type(self):
        """
        Get the setback feature type (required).
        must be one of the keys of `SETBACKS`
        """
        feature_type = self['feature_type']
        msg = ("feature_type must be one of: {}; got {}"
               .format(SETBACKS, feature_type))
        assert feature_type in SETBACKS, msg

        return feature_type

    @property
    def excl_fpath(self):
        """Get the exclusions .h5 file path (required)."""
        return self['excl_fpath']

    @property
    def features_path(self):
        """Get path to features file or directory (required)."""
        return self['features_path']

    @property
    def base_setback_dist(self):
        """Get plant height."""
        return self.get('base_setback_dist', None)

    @property
    def hub_height(self):
        """Get turbine hub height."""
        return self.get('hub_height', None)

    @property
    def rotor_diameter(self):
        """Get turbine rotor diameter."""
        return self.get('rotor_diameter', None)

    @property
    def regs_fpath(self):
        """Get regulations .csv path"""
        return self.get('regs_fpath', None)

    @property
    def multiplier(self):
        """Get generic max-tip height multiplier"""
        return self.get('multiplier', None)

    @property
    def replace(self):
        """Get replace flag"""
        return self.get('replace', False)

    @property
    def hsds(self):
        """Get hsds flag"""
        return self.get('hsds', False)

    @property
    def weights_calculation_upscale_factor(self):
        """Get upscale factor for weights calculation. """
        return self.get("weights_calculation_upscale_factor", None)

    @property
    def out_layers(self):
        """Get out_layers dictionary. """
        return self.get("out_layers", None)



# -*- coding: utf-8 -*-
"""
Abstract generic+local regulations
"""
from abc import ABC, abstractmethod
import logging
import geopandas as gpd

from rex.utilities import parse_table


logger = logging.getLogger(__name__)


class AbstractBaseRegulations(ABC):
    """ABC for county regulation values. """

    REQUIRED_COLUMNS = ["Feature Type", "Value Type", "Value", "FIPS"]

    def __init__(self, generic_regulation_value=None, regulations_fpath=None):
        """
        Parameters
        ----------
        generic_regulation_value : float | int | None, optional
            A generic regulation value to be applied where local
            regulations and/or ordinances are not given. A `None` value
            signifies that no regulation should be applied for regions
            without a local regulation. By default `None`.
        regulations_fpath : str | None, optional
            Path to regulations .csv or .gpkg file. At a minimum, this
            file must contain the following columns: `Feature Type`
            which labels the type of regulation that each row
            represents, `Value Type`, which specifies the type of the
            value (e.g. a multiplier or static height, etc.), `Value`,
            which specifies the numeric value of the regulation, and
            `FIPS`, which specifies a unique 5-digit code for each
            county (this can be an integer - no leading zeros required).
            A `None` value signifies that no local regulations should
            be applied. By default `None`.
        """

        self._generic_regulation_value = generic_regulation_value
        self._regulations_df = None
        self._preflight_check(regulations_fpath)

    def _preflight_check(self, regulations_fpath):
        """Apply preflight checks to the regulations path and multiplier.

        Run preflight checks on setback inputs:
        1) Ensure either a regulations .csv or a generic regulation
           value (or both) is provided
        2) Ensure regulations has county FIPS, map regulations to county
           geometries from exclusions .h5 file

        Parameters
        ----------
        regulations_fpath : str | None
            Path to regulations .csv file, if `None`, create global
            setbacks.
        """
        if regulations_fpath:
            try:
                self.df = parse_table(regulations_fpath)
            except ValueError:
                self.df = gpd.read_file(regulations_fpath)
            logger.debug('Found regulations provided in: {}'
                         .format(regulations_fpath))

        if (regulations_fpath is None
            and self._generic_regulation_value is None):
            msg = ('Regulations require a local regulation.csv file '
                   'and/or a generic regulation value!')
            logger.error(msg)
            raise RuntimeError(msg)

    @property
    def generic(self):
        """float | None: Regulation value used for global regulations. """
        return self._generic_regulation_value

    @property
    def df(self):
        """geopandas.GeoDataFrame | None: Regulations table. """
        return self._regulations_df

    @df.setter
    def df(self, regulations_df):
        if regulations_df is None:
            msg = "Cannot set df to `None`"
            logger.error(msg)
            raise ValueError(msg)
        self._regulations_df = regulations_df
        self._validate_regulations()

    def _validate_regulations(self):
        """Perform several validations on regulations"""

        self._convert_cols_to_title()
        self._check_for_req_missing_cols()
        self._remove_nans_from_req_cols()
        self._casefold(cols=['Feature Type', 'Value Type'])

    def _convert_cols_to_title(self):
        """Convert column names in regulations DataFrame to str.title(). """
        new_col_names = {col: col.lower().title()
                         for col in self._regulations_df.columns
                         if col.lower() not in {"geometry", "fips"}}
        self._regulations_df = self._regulations_df.rename(new_col_names,
                                                           axis=1)

    def _check_for_req_missing_cols(self):
        """Check for missing (required) columns in regulations DataFrame. """
        missing = [col for col in self.REQUIRED_COLUMNS
                   if col not in self._regulations_df]
        if any(missing):
            msg = ('Regulations are missing the following required columns: {}'
                   .format(missing))
            logger.error(msg)
            raise RuntimeError(msg)

    def _remove_nans_from_req_cols(self):
        """Remove rows with NaN values from required columns. """
        for col in self.REQUIRED_COLUMNS:
            na_rows = self._regulations_df[col].isna()
            self._regulations_df = self._regulations_df[~na_rows]

    def _casefold(self, cols):
        """Casefold column values. """
        for col in cols:
            vals = self._regulations_df[col].str.strip().str.casefold()
            self._regulations_df[col] = vals

    @property
    def locals_exist(self):
        """bool: Flag indicating wether local regulations exist. """
        return (self.df is not None and not self.df.empty)

    @property
    def generic_exists(self):
        """bool: Flag indicating wether generic regulations exist. """
        return self.generic is not None

    def __iter__(self):
        if self._regulations_df is None:
            return
        for ind, county_regulations in self.df.iterrows():
            regulation = self._county_regulation_value(county_regulations)
            if regulation is None:
                continue
            yield regulation, self.df.iloc[[ind]].copy()

    @abstractmethod
    def _county_regulation_value(self, county_regulations):
        """Retrieve county regulation value. """
        raise NotImplementedError




# -*- coding: utf-8 -*-
"""
reVX Utilities
"""
import geopandas as gpd
import numpy as np
import pyproj
import rasterio
import shapely
import skimage
import sklearn
from sklearn.metrics.pairwise import haversine_distances

from reV.utilities import log_versions as reV_log_versions
from reVX.version import __version__


STATES_ABBR_MAP = {
    'Alaska': 'AK',
    'Alabama': 'AL',
    'Arkansas': 'AR',
    'Arizona': 'AZ',
    'California': 'CA',
    'Colorado': 'CO',
    'Connecticut': 'CT',
    'District Of Columbia': 'DC',
    'Delaware': 'DE',
    'Florida': 'FL',
    'Georgia': 'GA',
    'Hawaii': 'HI',
    'Idaho': 'ID',
    'Illinois': 'IL',
    'Indiana': 'IN',
    'Iowa': 'IA',
    'Kansas': 'KS',
    'Kentucky': 'KY',
    'Louisiana': 'LA',
    'Massachusetts': 'MA',
    'Maryland': 'MD',
    'Maine': 'ME',
    'Michigan': 'MI',
    'Minnesota': 'MN',
    'Missouri': 'MO',
    'Mississippi': 'MS',
    'Montana': 'MT',
    'Nebraska': 'NE',
    'New Hampshire': 'NH',
    'New Jersey': 'NJ',
    'Nevada': 'NV',
    'New Mexico': 'NM',
    'New York': 'NY',
    'North Carolina': 'NC',
    'North Dakota': 'ND',
    'Ohio': 'OH',
    'Oklahoma': 'OK',
    'Oregon': 'OR',
    'Pennsylvania': 'PA',
    'Rhode Island': 'RI',
    'South Carolina': 'SC',
    'South Dakota': 'SD',
    'Tennessee': 'TN',
    'Texas': 'TX',
    'Utah': 'UT',
    'Virginia': 'VA',
    'Washington': 'WA',
    'Wisconsin': 'WI',
    'West Virginia': 'WV',
    'Wyoming': 'WY'
}


def coordinate_distance(coords1, coords2):
    """
    Compute the haversine distance between the two sets of coordinates.
    Results are in km

    Parameters
    ----------
    coords1 : ndarray
        First set of (lat, lon) coordinates
    coords2 : ndarray
        Second set of (lat, lon) coordinates

    Returns
    -------
    dist : ndarray
        Vector of haversine distances between coordinate set 1 and set 2 in km
    """
    dist = haversine_distances(np.radians(coords1), np.radians(coords2))
    if len(coords1) == 1:
        dist = dist.ravel()
    else:
        dist = np.diag(dist)

    # radius of the earth in kilometers # radius of the earth in km
    R = 6371.0

    return dist * R


def log_versions(logger):
    """Log package versions:
    - rex, reV, and reVX to info
    - h5py, numpy, pandas, scipy, PySAM, geopandas, pyproj, rasterio, shapely,
      sklearn, and skimage to debug

    Parameters
    ----------
    logger : logging.Logger
        Logger object to log memory message to.
    """
    logger.info('Running with reVX version {}'.format(__version__))
    reV_log_versions(logger)
    logger.debug('- geopandas version {}'.format(gpd.__version__))
    logger.debug('- pyproj version {}'.format(pyproj.__version__))
    logger.debug('- rasterio version {}'.format(rasterio.__version__))
    logger.debug('- shapely version {}'.format(shapely.__version__))
    logger.debug('- scikit-image version {}'.format(skimage.__version__))
    logger.debug('- scikit-learn version {}'.format(sklearn.__version__))



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
import os
from abc import abstractmethod
from warnings import warn
from itertools import product
import logging
import pathlib
import numpy as np
import geopandas as gpd
from rasterio import features as rio_features

from rex.utilities import log_mem
from reV.handlers.exclusions import ExclusionLayers
from reVX.utilities.exclusions import AbstractBaseExclusionsMerger

logger = logging.getLogger(__name__)


def features_with_centroid_in_county(features, cnty):
    """Find features with centroids within the given county.

    Parameters
    ----------
    features : geopandas.GeoDataFrame
        Features to setback from.
    cnty : geopandas.GeoDataFrame
        Regulations for a single county.

    Returns
    -------
    features : geopandas.GeoDataFrame
        Features that have centroid in county.
    """

    mask = features.centroid.within(cnty['geometry'].values[0])
    return features.loc[mask]


def features_clipped_to_county(features, cnty):
    """Clip features to the given county geometry.

    Parameters
    ----------
    features : geopandas.GeoDataFrame
        Features to setback from.
    cnty : geopandas.GeoDataFrame
        Regulations for a single county.

    Returns
    -------
    features : geopandas.GeoDataFrame
        Features clipped to county geometry.
    """
    tmp = gpd.clip(features, cnty)
    return tmp[~tmp.is_empty]


def _parse_excl_properties(excl_fpath, hsds=False):
    """Parse shape, chunk size, and profile from exclusions file.

    Parameters
    ----------
    excl_fpath : str
        Path to .h5 file containing exclusion layers, will also be
        the location of any new setback layers
    hsds : bool, optional
        Boolean flag to use h5pyd to handle .h5 'files' hosted on
        AWS behind HSDS. By default `False`.

    Returns
    -------
    shape : tuple
        Shape of exclusions datasets
    profile : str
        GeoTiff profile for exclusions datasets
    """
    with ExclusionLayers(excl_fpath, hsds=hsds) as exc:
        dset_shape = exc.shape
        profile = exc.profile

    if len(dset_shape) < 3:
        dset_shape = (1, ) + dset_shape

    logger.debug('Exclusions properties:\n'
                 'shape : {}\n'
                 'profile : {}\n'
                 .format(dset_shape, profile))

    return dset_shape, profile


class Rasterizer:
    """Helper class to rasterize shapes."""

    def __init__(self, excl_fpath, weights_calculation_upscale_factor,
                 hsds=False):
        """
        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing template layers. The raster will
            match the shape and profile of these layers.
        weights_calculation_upscale_factor : int
            If this value is an int > 1, the output will be a layer with
            **inclusion** weight values (floats ranging from 0 to 1).
            Note that this is backwards w.r.t the typical output of
            exclusion integer values (1 for excluded, 0 otherwise).
            Values <= 1 will still return a standard exclusion mask.
            For example, a cell that was previously excluded with a
            a boolean mask (value of 1) may instead be converted to an
            inclusion weight value of 0.75, meaning that 75% of the area
            corresponding to that point should be included (i.e. the
            exclusion feature only intersected a small portion - 25% -
            of the cell). This percentage inclusion value is calculated
            by upscaling the output array using this input value,
            rasterizing the exclusion features onto it, and counting the
            number of resulting sub-cells excluded by the feature. For
            example, setting the value to `3` would split each output
            cell into nine sub-cells - 3 divisions in each dimension.
            After the feature is rasterized on this high-resolution
            sub-grid, the area of the non-excluded sub-cells is totaled
            and divided by the area of the original cell to obtain the
            final inclusion percentage. Therefore, a larger upscale
            factor results in more accurate percentage values. However,
            this process is memory intensive and scales quadratically
            with the upscale factor. A good way to estimate your minimum
            memory requirement is to use the following formula:

            .. math:: memory (GB) = s_0 * s_1 * (sf^2 * 2 + 4) / 1073741824,

            where :math:`s_0` and :math:`s_1` are the dimensions (shape)
            of your exclusion layer and :math:`sf` is the scale factor
            (be sure to add several GB for any other overhead required
            by the rest of the process). If `None` (or a value <= 1),
            this process is skipped and the output is a boolean
            exclusion mask. By default `None`.
        """
        props = _parse_excl_properties(excl_fpath, hsds=hsds)
        self._shape, self._profile = props
        self._scale_factor = weights_calculation_upscale_factor or 1
        self._scale_factor = int(self._scale_factor // 1)

    @property
    def profile(self):
        """Geotiff profile.

        Returns
        -------
        dict
        """
        return self._profile

    @property
    def arr_shape(self):
        """Rasterize array shape.

        Returns
        -------
        tuple
        """
        return self._shape

    @property
    def inclusions(self):
        """Flag indicating wether or not the output raster represents
        inclusion values.

        Returns
        -------
        bool
        """
        return self._scale_factor > 1

    def _no_exclusions_array(self, multiplier=1):
        """Get an array of the correct shape representing no exclusions.

        The array contains all zeros, and a new one is created
        for every function call.

        Parameters
        ----------
        multiplier : int, optional
            Integer multiplier value used to scale up the dimensions of
            the array exclusions array (e.g. multiplier of 3 turns an
            array of shape (10, 20) into an array of shape (30, 60)).

        Returns
        -------
        np.array
            Array of zeros representing no exclusions.
        """
        high_res_shape = tuple(x * multiplier for x in self.arr_shape[1:])
        return np.zeros(high_res_shape, dtype='uint8')

    def rasterize(self, shapes):
        """Convert geometries into exclusions array.

        Parameters
        ----------
        shapes : list, optional
            List of geometries to rasterize (i.e. list(gdf["geometry])).
            If `None` or empty list, returns array of zeros.

        Returns
        -------
        arr : ndarray
            Rasterized array of shapes.
        """
        logger.debug('Generating exclusion array of shape {}'
                     .format(self.arr_shape))
        log_mem(logger)

        shapes = shapes or []
        shapes = [(geom, 1) for geom in shapes if geom is not None]

        if self.inclusions:
            return self._rasterize_to_weights(shapes)

        return self._rasterize_to_mask(shapes)

    def _rasterize_to_weights(self, shapes):
        """Rasterize features to weights using a high-resolution array."""

        if not shapes:
            return 1 - self._no_exclusions_array().astype(np.float32)

        hr_arr = self._no_exclusions_array(multiplier=self._scale_factor)
        new_transform = list(self.profile['transform'])[:6]
        new_transform[0] = new_transform[0] / self._scale_factor
        new_transform[4] = new_transform[4] / self._scale_factor

        rio_features.rasterize(shapes=shapes,
                               out=hr_arr,
                               out_shape=hr_arr.shape[1:],
                               fill=0,
                               transform=new_transform)

        arr = self._aggregate_high_res(hr_arr)
        return 1 - (arr / self._scale_factor ** 2)

    def _rasterize_to_mask(self, shapes):
        """Rasterize features with to an exclusion mask."""

        arr = self._no_exclusions_array()
        if shapes:
            rio_features.rasterize(shapes=shapes,
                                   out=arr,
                                   out_shape=arr.shape[1:],
                                   fill=0,
                                   transform=self.profile['transform'])

        return arr

    def _aggregate_high_res(self, hr_arr):
        """Aggregate the high resolution exclusions array to output shape. """

        arr = self._no_exclusions_array().astype(np.float32)
        for i, j in product(range(self._scale_factor),
                            range(self._scale_factor)):
            arr += hr_arr[i::self._scale_factor, j::self._scale_factor]
        return arr


class AbstractBaseSetbacks(AbstractBaseExclusionsMerger):
    """
    Create exclusions layers for setbacks
    """

    def __init__(self, excl_fpath, regulations, features, hsds=False,
                 weights_calculation_upscale_factor=None):
        """
        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new setback layers
        regulations : `~reVX.setbacks.regulations.SetbackRegulations`
            A `SetbackRegulations` object used to extract setback
            distances.
        features : str
            Path to file containing features to compute exclusions from.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        weights_calculation_upscale_factor : int, optional
            If this value is an int > 1, the output will be a layer with
            **inclusion** weight values instead of exclusion booleans.
            For example, a cell that was previously excluded with a
            a boolean mask (value of 1) may instead be converted to an
            inclusion weight value of 0.75, meaning that 75% of the area
            corresponding to that point should be included (i.e. the
            exclusion feature only intersected a small portion - 25% -
            of the cell). This percentage inclusion value is calculated
            by upscaling the output array using this input value,
            rasterizing the exclusion features onto it, and counting the
            number of resulting sub-cells excluded by the feature. For
            example, setting the value to `3` would split each output
            cell into nine sub-cells - 3 divisions in each dimension.
            After the feature is rasterized on this high-resolution
            sub-grid, the area of the non-excluded sub-cells is totaled
            and divided by the area of the original cell to obtain the
            final inclusion percentage. Therefore, a larger upscale
            factor results in more accurate percentage values. However,
            this process is memory intensive and scales quadratically
            with the upscale factor. A good way to estimate your minimum
            memory requirement is to use the following formula:

            .. math:: memory (GB) = s_0 * s_1 * (sf^2 * 2 + 4) / 1073741824,

            where :math:`s_0` and :math:`s_1` are the dimensions (shape)
            of your exclusion layer and :math:`sf` is the scale factor
            (be sure to add several GB for any other overhead required
            by the rest of the process). If `None` (or a value <= 1),
            this process is skipped and the output is a boolean
            exclusion mask. By default `None`.
        """
        self._rasterizer = Rasterizer(excl_fpath,
                                      weights_calculation_upscale_factor, hsds)
        super().__init__(excl_fpath, regulations, features, hsds)

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_fpath)
        return msg

    def parse_features(self):
        """Method to parse features.

        Returns
        -------
        `geopandas.GeoDataFrame`
            Geometries of features to setback from in exclusion
            coordinate system.
        """
        logger.debug("Loading features from {}".format(self._features))
        return (gpd.read_file(self._features)
                .to_crs(crs=self.profile['crs']))

    @property
    def description(self):
        """str: Description to be added to excl H5."""
        return ('{} computed with a base setback distance of {} and a '
                'multiplier of {} for a total generic setback value of {} '
                '(local exclusions may differ).'
                .format(self.__class__,
                        self._regulations.base_setback_dist,
                        self._regulations.multiplier,
                        self._regulations.generic))

    @property
    def no_exclusions_array(self):
        """np.array: Array representing no exclusions. """
        return self._rasterizer.rasterize(shapes=None)

    @property
    def exclusion_merge_func(self):
        """callable: Function to merge overlapping exclusion layers. """
        return np.minimum if self._rasterizer.inclusions else np.maximum

    def pre_process_regulations(self):
        """Reduce regulations to state corresponding to features.

        """
        mask = self._regulation_table_mask()
        if not mask.any():
            msg = "Found no local regulations!"
            logger.warning(msg)
            warn(msg)

        self._regulations.df = (self.regulations_table[mask]
                                .reset_index(drop=True))
        logger.debug('Computing setbacks for regulations in {} counties'
                     .format(len(self.regulations_table)))

    def _local_exclusions_arguments(self, __, cnty):
        """Compile and return arguments to `compute_local_exclusions`. """
        idx = self.features.sindex.intersection(cnty.total_bounds)
        cnty_features = self.features.iloc[list(idx)].copy()
        return cnty_features, self._feature_filter, self._rasterizer

    @staticmethod
    def compute_local_exclusions(regulation_value, cnty, *args):
        """Compute local features setbacks.

        This method will compute the setbacks using a county-specific
        regulations file that specifies either a static setback or a
        multiplier value that will be used along with the base setback
        distance to compute the setback.

        Parameters
        ----------
        regulation_value : float | int
            Setback distance in meters.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.
        features : geopandas.GeoDataFrame
            Features for the local county.
        feature_filter : callable
            A callable function that takes `features` and `cnty` as
            inputs and outputs a geopandas.GeoDataFrame with features
            clipped and/or localized to the input county.
        rasterizer : Rasterizer
            Instance of `Rasterizer` class used to rasterize the
            buffered county features.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        features, feature_filter, rasterizer = args
        logger.debug('- Computing setbacks for county FIPS {}'
                     .format(cnty.iloc[0]['FIPS']))
        log_mem(logger)
        features = feature_filter(features, cnty)
        features = list(features.buffer(regulation_value))
        return rasterizer.rasterize(features)

    # pylint: disable=arguments-differ
    def compute_generic_exclusions(self, **__):
        """Compute generic setbacks.

        This method will compute the setbacks using a generic setback
        of `base_setback_dist * multiplier`.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        logger.info('Computing generic setbacks')
        generic_regs_dne = (self._regulations.generic is None
                            or np.isclose(self._regulations.generic, 0))
        if generic_regs_dne:
            return self.no_exclusions_array

        setbacks = list(self.features.buffer(self._regulations.generic))

        return self._rasterizer.rasterize(setbacks)

    @classmethod
    def input_output_filenames(cls, out_dir, features_fpath, *__, **___):
        """Generate pairs of input/output file names.

        Parameters
        ----------
        out_dir : str
            Path to output file directory.
        features_fpath : str
            Path to features file. This path can contain
            any pattern that can be used in the glob function.
            For example, `/path/to/features/[A]*` would match
            with all the features in the directory
            `/path/to/features/` that start with "A". This input
            can also be a directory, but that directory must ONLY
            contain feature files. If your feature files are mixed
            with other files or directories, use something like
            `/path/to/features/*.geojson`.
        kwargs : dict
            Dictionary of extra keyword-argument pairs used to
            instantiate the `exclusion_class`.

        Yields
        ------
        tuple
            An input-output filename pair.
        """
        for fpath in cls.get_feature_paths(features_fpath):
            fn = os.path.basename(fpath)
            geotiff = ".".join(fn.split('.')[:-1] + ['tif'])
            yield fpath, os.path.join(out_dir, geotiff)

    @staticmethod
    def get_feature_paths(features_fpath):
        """Ensure features path exists and return as list.

        Parameters
        ----------
        features_fpath : str
            Path to features file. This path can contain
            any pattern that can be used in the glob function.
            For example, `/path/to/features/[A]*` would match
            with all the features in the directory
            `/path/to/features/` that start with "A". This input
            can also be a directory, but that directory must ONLY
            contain feature files. If your feature files are mixed
            with other files or directories, use something like
            `/path/to/features/*.geojson`.

        Returns
        -------
        features_fpath : list
            Features path as a list of strings.

        Notes
        -----
        This method is required for `run` classmethods for
        feature setbacks that are spread out over multiple
        files.
        """
        glob_path = pathlib.Path(features_fpath)
        if glob_path.is_dir():
            glob_path = glob_path / '*'

        paths = [str(f) for f in glob_path.parent.glob(glob_path.name)]
        if not paths:
            msg = 'No files found matching the input {!r}!'
            msg = msg.format(features_fpath)
            logger.error(msg)
            raise FileNotFoundError(msg)

        return paths

    @staticmethod
    def _feature_filter(features, cnty):
        """Filter the features given a county."""
        return features_with_centroid_in_county(features, cnty)

    @abstractmethod
    def _regulation_table_mask(self):
        """Return the regulation table mask for setback feature. """
        raise NotImplementedError



"""
Various utility functions to prep data for AOSWT processing.

Mike Bannister 5/2022
"""
import os
import json
import logging
from functools import reduce

import h5py
import numpy as np
import pandas as pd
import geopandas as gpd
import rasterio as rio
from rasterio import features
from rasterio.warp import reproject, Resampling
from shapely.geometry import Point, LineString

import rex

logger = logging.getLogger(__name__)


def _sum(a, b):
    return a + b


def convert_pois_to_lines(poi_csv_f, template_f, out_f):
    """
    Convert POIs in CSV to lines and save in a geopackage as substations. Also
    create a fake transmission line to connect to the substations.

    Parameters
    ----------
    poi_csv_f : str
        Path to CSV file with POIs in it
    template_f : str
        Path to template raster with CRS to use for geopackage
    out_f : str
        Path and file name for geopackage
    """
    logger.info('Converting POIs in {} to lines in {}'
                .format(poi_csv_f, out_f))
    with rio.open(template_f) as ras:
        crs = ras.crs

    df = pd.read_csv(poi_csv_f)[['POI Name', 'State', 'Voltage (kV)', 'Lat',
                                 'Long']]

    pts = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.Long, df.Lat))
    pts = pts.set_crs('EPSG:4326')
    pts = pts.to_crs(crs)

    # Convert points to short lines
    new_geom = []
    for pt in pts.geometry:
        end = Point(pt.x + 50, pt.y + 50)
        line = LineString([pt, end])
        new_geom.append(line)
    lines = pts.set_geometry(new_geom, crs=crs)

    # Append some fake values to make the LCP code happy
    lines['ac_cap'] = 9999999
    lines['category'] = 'Substation'
    lines['voltage'] = 500  # kV
    lines['trans_gids'] = '[9999]'

    # add a fake trans line for the subs to connect to to make LCP code happy
    trans_line = pd.DataFrame(
        {
            'POI Name': 'fake',
            'ac_cap': 9999999,
            'category': 'TransLine',
            'voltage': 500,  # kV
            'trans_gids': None
        },
        index=[9999]
    )

    trans_line = gpd.GeoDataFrame(trans_line)
    geo = LineString([Point(0, 0), Point(100000, 100000)])
    trans_line = trans_line.set_geometry([geo], crs=crs)

    pois = lines.append(trans_line)
    pois['gid'] = pois.index

    pois.to_file(out_f, driver="GPKG")
    logger.info('Complete')


class CombineRasters:
    """
    Combine layers to create composite friction and barrier rasters. Merge
    with existing land cost and barriers and save to h5.
    """
    OFFSHORE_FRICTION_FNAME = 'offshore_friction.tif'
    OFFSHORE_BARRIERS_FNAME = 'offshore_barriers.tif'
    COMBO_LAYER_FNAME = 'combo_{}.tif'
    LAND_MASK_FNAME = 'land_mask.tif'
    SLOPE_CUTOFF = 15  # slopes >= this value are barriers

    def __init__(self, template_f, layer_dir=''):
        """
        Parameters
        ----------
        template_f : str
            Path to template raster with CRS to use for geopackage
        layer_dir : str, optional
            Directory to prepend to barrier and friction layer filenames
        """
        self.layer_dir = layer_dir

        self._os_profile = self._extract_profile(template_f)
        self._os_profile['dtype'] = ('MUST SET in {}.profile()!'
                                     .format(self.__class__.__name__))
        self._os_shape = (self.profile()['height'],
                          self.profile()['width'])

        self._os_barriers = None  # (uint8) off-shore barrier raster
        self._os_friction = None  # (float32) off-shore friction raster
        self._land_mask = None  # (bool) land mask raster, true indicates land

    def create_land_mask(self, mask_shp_f, save_tiff=False):
        """
        Create the land mask layer from a vector file

        Parameters
        ----------
        mask_shp_f : str
            Full path to mask gpgk or shp file
        save_tiff : bool
            Save mask as tiff if true
        """
        ldf = gpd.read_file(mask_shp_f)
        logger.info('Rasterizing {}'.format(mask_shp_f))
        l_geom = list(ldf.geometry)
        l_rast = features.rasterize(l_geom, out_shape=self._os_shape, fill=0,
                                    out=None,
                                    transform=self.profile()['transform'],
                                    all_touched=False, default_value=1,
                                    dtype=None)

        if save_tiff:
            logger.info('Saving land mask to {}'.format(self.LAND_MASK_FNAME))
            self._save_tiff(l_rast, self.LAND_MASK_FNAME)

        self._land_mask = l_rast == 1
        logger.info('Rasterizing complete')

    def load_land_mask(self, mask_f=None):
        """
        Load the land mask layer from a tiff. This does not need to be called
        if self.create_land_mask() was run previously.

        Parameters
        ----------
        mask_f : str
            Full path to mask tiff
        """
        if mask_f is None:
            mask_f = self.LAND_MASK_FNAME
        with rio.open(mask_f) as ras:
            l_rast = ras.read(1)

        assert l_rast.max() == 1
        assert l_rast.shape == self._os_shape

        self._land_mask = l_rast == 1
        logger.info('Successfully loaded land mask from {}'.format(mask_f))

    def build_off_shore_friction(self, friction_files, slope_file=None,
                                 save_tiff=None):
        """
        Combine off-shore friction layers.

        friction_files : list of tuples
            Friction layers to combine and raster value to friction dict.
            Tuples are in the format:

                ({ras_val1: fric_val1, ras_val2: fric_val2, ...}, 'fname.tif')

            where ras_valX is the value in the raster, and fric_valX is the
            desired friction values. Any values in the raster that are not
            specified in the dict are assumed to have no friction. 'fname.tif'
            is the file name of the raster.
        slope_file : str, optional
            Path to slope friction tiff
        save_tiff : bool, optional
            Save composite friction to tiff if true
        """
        logger.info('Loading friction layers')
        fr_layers = {}
        for fr_dict, f in friction_files:
            d = None
            for k, val in fr_dict.items():
                logger.info('--- setting raster value {} to friction '
                            '{} for {}'.format(k, val, f))
                tmp_d = self._load_layer(f, k) * val
                d = tmp_d if d is None else d + tmp_d

            assert d.shape == self._os_shape and d.min() == 0
            fr_layers[f] = d.astype('uint16')

        if slope_file is not None:
            logger.info('--- calculating slope friction')
            HIGH_FRICTION = 10
            MEDIUM_FRICTION = 5
            LOW_FRICTION = 1

            if not os.path.exists(slope_file):
                slope_file = os.path.join(self.layer_dir, slope_file)
            if not os.path.exists(slope_file):
                raise FileNotFoundError('Unable to find {}'.format(slope_file))

            d = rio.open(slope_file).read(1)
            d[d < 0] = 0
            assert d.shape == self._os_shape and d.min() == 0
            # Slope >= SLOPE_CUTOFF is also included in barriers
            d2 = np.where(d >= self.SLOPE_CUTOFF, HIGH_FRICTION, d)
            d2 = np.where(d < self.SLOPE_CUTOFF, MEDIUM_FRICTION, d2)
            d2 = np.where(d < 10, LOW_FRICTION, d2)

            fr_layers[slope_file] = d2.astype('uint16')

        logger.info('Combining off-shore friction layers')
        self._os_friction = reduce(_sum, fr_layers.values()).astype('uint16')

        if save_tiff:
            logger.info('Saving combined friction to tiff')
            self._save_tiff(self._os_friction, self.OFFSHORE_FRICTION_FNAME)

        logger.info('Done processing friction layers')

    def build_off_shore_barriers(self, barrier_files, fi_files,
                                 slope_file=None, save_tiff=False):
        """
        Combine off-shore barrier layers

        Parameters
        ----------
        barrier_files : list of tuples (int|list, str)
            Barrier layers to combine. Tuples are in one of two formats:
                (X, 'fname.tif') or
                ([X1, X2, ...], 'fname.tif')
            Where 'fname.tif' is the raster file, and X is the raster value
            to use as the barrier. Alternatively, a list of multiple values can
            be used as barriers. Any other values in the raster are assumed
            to be open to tranmission.
        fi_files : list of tuples (int, str)
            Force include layers. These will override the barrier layers.
            Tuple format is the same as for barrier_files, however the specifed
            raster values are force included.
        slope_file : str, optional
            Path to slope tiff
        save_tiff : bool, options
            Save composite layer to geotiff if true
        """
        logger.info('Loading barrier layers')
        barrier_layers = {}
        for val, f in barrier_files:
            logger.info('--- {}'.format(f))
            d = self._load_layer(f, val)
            assert d.shape == self._os_shape and d.min() == 0 and d.max() == 1
            barrier_layers[f] = d

        if slope_file is not None:
            logger.info('--- calculating slope barrier')
            if not os.path.exists(slope_file):
                slope_file = os.path.join(self.layer_dir, slope_file)
            if not os.path.exists(slope_file):
                raise FileNotFoundError('Unable to find {}'.format(slope_file))

            d = rio.open(slope_file).read(1)
            assert d.shape == self._os_shape
            d2 = np.where(d < self.SLOPE_CUTOFF, 0, d)
            d2 = np.where(d >= self.SLOPE_CUTOFF, 1, d2)

            barrier_layers[slope_file] = d2.astype('uint8')

        # Add all the exclusion layers together and normalize
        logger.info('Building composite off-shore barrier layers')
        comp_bar = reduce(_sum, barrier_layers.values())
        comp_bar[comp_bar >= 1] = 1

        if len(fi_files) > 0:
            logger.info('Loading forced inclusion layers')
            fi_layers = {}
            for val, f in fi_files:
                logger.info('--- {}'.format(f))
                d = self._load_layer(f, val)
                assert d.shape == self._os_shape and d.min() == 0 and \
                    d.max() == 1
                fi_layers[f] = d

            logger.info('Building composite forced inclusion layers')
            comp_fi = reduce(_sum, fi_layers.values())
            comp_fi[comp_fi >= 1] = 1

            # Subtract fi from barriers
            comp_bar = comp_bar.astype('int8') - comp_fi.astype('int8')
            comp_bar[comp_bar < 0] = 0

        assert comp_bar.max() == 1
        assert comp_bar.min() == 0

        if save_tiff:
            logger.info('Saving barriers to {}'
                        .format(self.OFFSHORE_BARRIERS_FNAME))
            self._save_tiff(comp_bar, self.OFFSHORE_BARRIERS_FNAME)

        self._os_barriers = comp_bar
        logger.info('Done building barrier layers')

    def merge_os_and_land_friction(self, land_h5, land_cost_layer,
                                   aoswt_h5, os_friction_layer=None,
                                   os_friction_f=None, land_cost_mult=1,
                                   save_tiff=False):
        """
        Combine off-shore friction and land cost layers and save to h5. For
        land it's called cost for legacy reasons, and for off-shore it's called
        friction, but it's really the same thing.

        Parameters
        ----------
        land_h5 : str
            Path to h5 file w/ land barrier
        land_cost_layer : str
            Name of land barrier layer in h5 to use
        aoswt_h5 : str
            Path to h5 file to save combined barriers in
        os_friction_layer : str | None, optional
            Name for friction layer in off-shore h5. Use land_cost_layer
            if None.
        os_friction_f : str | None, optional
            Path to cached offshore friction raster. If None, will try to pull
            data from self._os_friction
        land_cost_mult : float, optional
            Multiplier for land costs
        save_tiff : bool, optional
            Save composite barrier layer to geotiff if true
        """
        # Try to load friction from self first, then look for tiff
        if self._os_friction is not None:
            os_friction = self._os_friction
        else:
            if os_friction_f is None:
                os_friction_f = self.OFFSHORE_FRICTION_FNAME
            if not os.path.exists(os_friction_f):
                msg = ('Off-shore friction has not been calculated and cached'
                       ' friction was not found at {}. Please run {}.'
                       'build_off_shore_friction() first or pass a valid '
                       'filename to os_friction_f'
                       .format(os_friction_f, self.__class__.__name__))
                raise AttributeError(msg)

            logger.info('Loading off-shore friction from {}'
                        .format(os_friction_f))
            with rio.open(os_friction_f) as ras:
                os_friction = ras.read(1)

        if os_friction_layer is None:
            os_friction_layer = land_cost_layer

        self._merge_os_and_land_layers(os_friction, land_h5, land_cost_layer,
                                       aoswt_h5, os_friction_layer,
                                       layer_name='friction',
                                       land_mult=land_cost_mult,
                                       save_tiff=save_tiff,
                                       dtype='float32')

    def merge_os_and_land_barriers(self, land_h5, land_barrier_layer,
                                   aoswt_h5, os_barrier_layer=None,
                                   os_barriers_f=None, save_tiff=False):
        """
        Combine off-shore and land barrier layers and save to h5

        Parameters
        ----------
        land_h5 : str
            Path to h5 file w/ land barrier
        land_barrier_layer : str
            Name of land barrier layer in h5 to use
        aoswt_h5 : str
            Path to h5 file to save combined barriers in
        os_barrier_layer : str | None
            Name for barrier layer in off-shore h5. Use land_barrier_layer
            if None.
        os_barriers_f : str | None, optional
            Path to offshore barrier raster. If None, will try to pull data
            from self._os_barriers
        save_tiff : bool, options
            Save composite barrier layer to geotiff if true
        """
        # Try to load barriers from self first, then look for tiff
        if self._os_barriers is not None:
            os_barriers = self._os_barriers
        else:
            if os_barriers_f is None:
                os_barriers_f = self.OFFSHORE_BARRIERS_FNAME
            if not os.path.exists(os_barriers_f):
                msg = ('Off-shore barriers have not been calculated and cached'
                       ' barriers were not found at {}. Please run {}.'
                       'build_off_shore_barriers() first or pass a valid '
                       'filename to os_barriers_f'
                       .format(os_barriers_f, self.__class__.__name__))
                raise AttributeError(msg)

            logger.info('Loading off-shore barriers from {}'
                        .format(os_barriers_f))
            with rio.open(os_barriers_f) as ras:
                os_barriers = ras.read(1)

        if os_barrier_layer is None:
            os_barrier_layer = land_barrier_layer

        self._merge_os_and_land_layers(os_barriers, land_h5,
                                       land_barrier_layer, aoswt_h5,
                                       os_barrier_layer, layer_name='barriers',
                                       save_tiff=save_tiff)

    def _merge_os_and_land_layers(self, os_data, land_h5, land_layer, aoswt_h5,
                                  os_layer, layer_name='data', land_mult=1,
                                  dtype='uint8', save_tiff=False,
                                  init_dest=-1):
        """
        Combine off-shore and land layers and save to h5

        Parameters
        ----------
        os_data : np.ndarray
            Off-shore data to merge with land data and save to h5
        land_h5 : str
            Path to h5 file w/ land layer
        land_layer : str
            Name of land layer in h5 to use
        aoswt_h5 : str
            Path to h5 file to save combined layer in
        os_layer : str
            Name for layer in off-shore h5.
        layer_name : str, optional
            Layer name for printing status, saving to tiff, and storing
            combined data on self
        land_mult : float, optional
            Multiplier for values in land layer
        dtype : str
            Data type to use for combined raster
        save_tiff : bool, options
            Save composite barrier layer to geotiff if true
        init_dest : int | float
            Initial value to seed combined raster
        """
        # Load land layer
        logger.info('Loading land {} "{}" from {}'
                    .format(layer_name, land_layer, land_h5))
        with rex.Resource(land_h5) as res:
            profile_json = res.attrs[land_layer]['profile']
            old_land_profile = json.loads(profile_json)
            old_land_data = res[land_layer][0]

        # Reproject land barriers to new off-shore projection
        logger.info('Reprojecting land {}'.format(layer_name))
        land_data = np.ones(self._os_shape, dtype=dtype)
        reproject(old_land_data,
                  destination=land_data,
                  src_transform=old_land_profile['transform'],
                  src_crs=old_land_profile['crs'],
                  dst_transform=self.profile()['transform'],
                  dst_crs=self.profile()['crs'],
                  dst_resolution=self._os_shape, num_threads=5,
                  resampling=Resampling.nearest,
                  INIT_DEST=init_dest)

        assert os_data.shape == land_data.shape
        setattr(self, '_land_{}'.format(layer_name), land_data)

        # Combine the land and off-shore data
        logger.info('Combining land and off-shore {}'.format(layer_name))
        combo_data = land_data * land_mult
        # pylint: disable=invalid-unary-operand-type
        combo_data[~self.land_mask] = os_data[~self.land_mask]
        combo_data = combo_data.astype(dtype)

        if save_tiff:
            fname = self.COMBO_LAYER_FNAME.format(layer_name)
            logger.info('Saving combined {} to {}'.format(layer_name, fname))
            self._save_tiff(combo_data, fname)

        setattr(self, '_combo_{}'.format(layer_name), combo_data)

        logger.info('Writing combined data to "{}" in {}'
                    .format(os_layer, aoswt_h5))
        combo_data = combo_data[np.newaxis, ...]
        with h5py.File(aoswt_h5, 'a') as f:
            if os_layer in f.keys():
                dset = f[os_layer]
                dset[...] = combo_data
            else:
                f.create_dataset(os_layer, data=combo_data)

    def create_aoswt_h5(self, aoswt_ex_h5, aoswt_h5, overwrite=False):
        """
        Create a new h5 file to save AOSWT data in.

        Parameters
        ----------
        aoswt_ex_h5 : str
            Path to existing h5 file w/ off-shore shape
        aoswt_h5 : str
            Path for new h5 file to create
        overwrite : bool, optional
            Overwrite existing h5 file if True

        """
        if os.path.exists(aoswt_h5) and not overwrite:
            raise AttributeError('File {} exits'.format(aoswt_h5))

        with rex.Resource(aoswt_ex_h5) as res:
            lats = res['latitude']
            lngs = res['longitude']
            global_attrs = res.global_attrs

        assert lats.shape == self._os_shape
        regions = np.ones(self._os_shape, dtype='uint8')

        with h5py.File(aoswt_h5, 'w') as f:
            f.create_dataset('longitude', data=lngs)
            f.create_dataset('latitude', data=lats)
            f.create_dataset('ISO_regions', data=regions[np.newaxis, ...],
                             dtype='uint8')
            for k, v in global_attrs.items():
                f.attrs[k] = v

    @property
    def land_mask(self):
        """np.ndarray: Land mask layer."""
        if self._land_mask is None:
            cls_name = self.__class__.__name__
            msg = ('Must run {0}.create_land_mask() or {0}.'
                   'load_land_mask() first'.format(cls_name))
            raise RuntimeError(msg)

        return self._land_mask

    def profile(self, dtype=None):
        """Copy CRS Profile.

        Parameters
        ----------
        dtype : str, optional
            Optional dtype fill. By default, `None`.

        Returns
        -------
        dict
            CRS Profile.
        """
        prof = self._os_profile.copy()
        if dtype:
            prof['dtype'] = dtype
        return prof

    @staticmethod
    def _extract_profile(template_f):
        """Extract profile from file. """
        with rio.open(template_f) as ras:
            profile = {
                'crs': ras.crs,
                'transform': ras.transform,
                'height': ras.height,
                'width': ras.width,
                # 'dtype': ras.dtype,
                'count': 1,
                'compress': 'lzw'
            }
        return profile

    def _load_layer(self, f, val, verbose=False):
        """
        Load a layer from a tiff and set appropriate cells to 1

        Parameters
        ----------
        f : str
            File to load
        val : int | list of int
            Value(s) in file to set to a value of 1. All other values are set
            to zero.

        Returns
        -------
        d : np.ndarray
            Raster layer
        """
        if not os.path.exists(f):
            f_old = f
            f = os.path.join(self.layer_dir, f)
        if not os.path.exists(f):
            raise FileNotFoundError('Unable to find file {} or {}'
                                    .format(f_old, f))

        name = f.split('/')[-1]
        if verbose:
            logger.info('Processing val {} for {}'.format(val, name))
        d = rio.open(f).read(1)
        if isinstance(val, int):
            d[d != val] = 0
            d[d == val] = 1
        elif isinstance(val, list):
            d[~np.in1d(d, val).reshape(d.shape)] = 0
            d[np.in1d(d, val).reshape(d.shape)] = 1
        else:
            raise AttributeError('Unknown type for val: {} - {}'
                                 .format(val, type(val)))
        if verbose:
            logger.info(d.shape, d.max(), d.min())
        return d

    def _save_tiff(self, data, f_name):
        """
        Save data to a geotiff

        Parameters
        ----------
            data : np.array
                Data to save
            f_name : str
                File name to save
        """
        dtype = data.dtype
        with rio.open(f_name, 'w', **self.profile(dtype=dtype)) as outf:
            outf.write(data, indexes=1)



# -*- coding: utf-8 -*-
"""
Module to compute least cost transmission paths, distances, AND costs
for one or more SC points
"""
import geopandas as gpd
import json
import logging
import numpy as np
import os
import pandas as pd
import rasterio
from pyproj.crs import CRS
from scipy.spatial import cKDTree
from shapely.geometry import Point
from concurrent.futures import as_completed
import time

from reV.handlers.exclusions import ExclusionLayers
from reV.supply_curve.extent import SupplyCurveExtent
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

from reVX.least_cost_xmission.config import (TRANS_LINE_CAT, LOAD_CENTER_CAT,
                                             SINK_CAT, SUBSTATION_CAT)
from reVX.least_cost_xmission.least_cost_paths import LeastCostPaths
from reVX.least_cost_xmission.trans_cap_costs import TransCapCosts

logger = logging.getLogger(__name__)


class LeastCostXmission(LeastCostPaths):
    """
    Compute Least Cost tie-line paths and full transmission cap cost
    for all possible connections to all supply curve points
    """
    REQUIRED_LAYRES = ['transmission_barrier', 'ISO_regions']

    def __init__(self, cost_fpath, features_fpath, resolution=128,
                 xmission_config=None, min_line_length=0):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        min_line_length : int | float, optional
            Minimum line length in km, by default 0.
        """
        self._check_layers(cost_fpath)
        self._config = TransCapCosts._parse_config(
            xmission_config=xmission_config)

        (self._sc_points, self._features,
         self._sub_lines_mapping, self._shape) =\
            self._map_to_costs(cost_fpath, features_fpath,
                               resolution=resolution)
        self._cost_fpath = cost_fpath
        self._tree = None
        self._sink_coords = None
        self._min_line_len = min_line_length

        logger.debug('{} initialized'.format(self))

    def __repr__(self):
        msg = ("{} to be computed for {} sc_points and {} features"
               .format(self.__class__.__name__,
                       len(self.sc_points),
                       len(self.features)))

        return msg

    @property
    def sc_points(self):
        """
        Table of supply curve points

        Returns
        -------
        gpd.GeoDataFrame
        """
        return self._sc_points

    @property
    def features(self):
        """
        Table of features to compute paths for

        Returns
        -------
        pandas.DataFrame
        """
        return self._features

    @property
    def sub_lines_mapping(self):
        """
        Series mapping substations  to the transmission lines connected
        to each substation

        Returns
        -------
        pandas.Series
        """
        return self._sub_lines_mapping

    @property
    def sink_coords(self):
        """
        Inf sink coordinates (row, col)

        Returns
        -------
        ndarray
        """
        if self._sink_coords is None:
            mask = self.features['category'] == SINK_CAT
            self._sink_coords = self.features.loc[mask, ['row', 'col']].values

        return self._sink_coords

    @property
    def sink_tree(self):
        """
        cKDTree for infinite sinks

        Returns
        -------
        cKDTree
        """
        if self._tree is None:
            self._tree = cKDTree(self.sink_coords)

        return self._tree

    @staticmethod
    def _load_trans_feats(features_fpath):
        """
        Load existing transmission features from disk. Substations will
        be loaded from cache file if it exists

        Parameters
        ----------
        features_fpath : str
            Path to GeoPackage with trans features

        Returns
        -------
        features : gpd.GeoDataFrame
            DataFrame of transmission features
        sub_line_map : pandas.Series
            Mapping of sub-station trans_gid to connected transmission
            line trans_gids
        """
        logger.debug('Loading transmission features')
        features = gpd.read_file(features_fpath)
        features = features.drop(columns=['bgid', 'egid', 'cap_left'],
                                 errors='ignore')
        mapping = {'gid': 'trans_gid', 'trans_gids': 'trans_line_gids'}
        features = features.rename(columns=mapping)

        features['min_volts'] = 0
        features['max_volts'] = 0

        # Transmission lines
        mask = features['category'] == TRANS_LINE_CAT
        voltage = features.loc[mask, 'voltage'].values
        features.loc[mask, 'min_volts'] = voltage
        features.loc[mask, 'max_volts'] = voltage

        # Load Center and Sinks
        mask = features['category'].isin([LOAD_CENTER_CAT, SINK_CAT])
        features.loc[mask, 'min_volts'] = 1
        features.loc[mask, 'max_volts'] = 9999

        sub_lines_map = {}
        mask = features['category'] == SUBSTATION_CAT
        bad_subs = np.zeros(len(features), dtype=bool)
        for idx, row in features.loc[mask].iterrows():
            gid = row['trans_gid']
            lines = row['trans_line_gids']
            if isinstance(lines, str):
                lines = json.loads(lines)

            sub_lines_map[gid] = lines
            lines_mask = features['trans_gid'].isin(lines)
            voltage = features.loc[lines_mask, 'voltage'].values

            if np.max(voltage) >= 69:
                features.loc[idx, 'min_volts'] = np.min(voltage)
                features.loc[idx, 'max_volts'] = np.max(voltage)
            else:
                bad_subs[idx] = True

        if any(bad_subs):
            msg = ("The following sub-stations do not have the minimum "
                   "required voltage of 69 kV and will be dropped:\n{}"
                   .format(features.loc[bad_subs, 'trans_gid']))
            logger.warning(msg)
            features = features.loc[~bad_subs].reset_index(drop=True)

        return features, pd.Series(sub_lines_map)

    @staticmethod
    def _create_sc_points(cost_fpath, resolution=128):
        """
        Load SC points, covert row/col to array wide, and determine x/y
        for reV projection

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        resolution : int, optional
            SC point resolution, by default 128

        Returns
        sc_points : gpd.GeoDataFrame
            SC points
        """
        logger.debug('Loading Supply Curve Points for res: {}'
                     .format(resolution))
        sce = SupplyCurveExtent(cost_fpath, resolution=resolution)
        sc_points = sce.points.rename(columns={'row_ind': 'sc_row_ind',
                                               'col_ind': 'sc_col_ind'})
        shape = sce.excl_shape
        sc_points['sc_point_gid'] = sc_points.index.values

        row = np.round(sc_points['sc_row_ind'] * resolution + resolution / 2)
        row = np.where(row >= shape[0], shape[0] - 1, row)
        sc_points['row'] = row.astype(int)

        col = np.round(sc_points['sc_col_ind'] * resolution + resolution / 2)
        col = np.where(col >= shape[1], shape[1] - 1, col)
        sc_points['col'] = col.astype(int)

        return sc_points

    @staticmethod
    def _get_feature_cost_indices(features, crs, transform, shape):
        """
        Map features to cost row, col indices using rasterio transform

        Parameters
        ----------
        features : gpd.GeoDataFrame
            GeoDataFrame of features to map to cost raster
        crs : pyproj.crs.CRS
            CRS of cost raster
        transform : raster.Affine
            Transform of cost raster
        shape : tuple
            Cost raster shape

        Returns
        -------
        row : ndarray
            Vector of row indices for each feature
        col : ndarray
            Vector of col indices for each features
        mask : ndarray
            Boolean mask of features with indices outside of cost raster
        """
        row, col, mask = super(LeastCostXmission,
                               LeastCostXmission)._get_feature_cost_indices(
            features, crs, transform, shape)

        t_lines = features['category'] == TRANS_LINE_CAT
        mask |= t_lines

        row[t_lines] = np.where(row[t_lines] >= 0, row[t_lines], 0)
        row[t_lines] = np.where(row[t_lines] < shape[0], row[t_lines],
                                shape[0] - 1)
        col[t_lines] = np.where(col[t_lines] >= 0, col[t_lines], 0)
        col[t_lines] = np.where(col[t_lines] < shape[1], col[t_lines],
                                shape[1] - 1)

        return row, col, mask

    @classmethod
    def _map_to_costs(cls, cost_fpath, features_fpath, resolution=128):
        """
        Map supply curve points and transmission features to cost array
        pixel indices

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128

        Returns
        -------
        sc_point : gpd.GeoDataFrame
            Table of supply curve points to connect to transmission
        features : gpd.GeoDataFrame
            Table of transmission features
        sub_lines_map : pandas.Series
            Series mapping substations  to the transmission lines
            connected to each substation
        """
        with ExclusionLayers(cost_fpath) as f:
            crs = CRS.from_string(f.crs)
            transform = rasterio.Affine(*f.profile['transform'])
            shape = f.shape
            regions = f['ISO_regions']

        features, sub_lines_map = cls._load_trans_feats(features_fpath)
        row, col, mask = cls._get_feature_cost_indices(features, crs,
                                                       transform, shape)
        if any(~mask):
            msg = ("The following features are outside of the cost exclusion "
                   "domain and will be dropped:\n{}"
                   .format(features.loc[~mask, 'trans_gid']))
            logger.warning(msg)
            row = row[mask]
            col = col[mask]
            features = features.loc[mask].reset_index(drop=True)

        features['row'] = row
        features['col'] = col
        features['region'] = regions[row, col]

        logger.debug('Converting SC points to GeoDataFrame')
        sc_points = cls._create_sc_points(cost_fpath, resolution=resolution)
        x, y = rasterio.transform.xy(transform, sc_points['row'].values,
                                     sc_points['col'].values)
        geo = [Point(xy) for xy in zip(x, y)]
        sc_points = gpd.GeoDataFrame(sc_points, crs=features.crs,
                                     geometry=geo)

        return sc_points, features, sub_lines_map, shape

    def _clip_to_sc_point(self, sc_point, tie_line_voltage, nn_sinks=2,
                          clipping_buffer=1.05, radius=None):
        """
        Clip costs raster to AOI around SC point, and get substations,
        load centers, and sinks within the clipped region.

        Parameters
        ----------
        sc_point : gpd.GeoSeries
            SC point to clip raster around
        nn_sinks : int, optional
            Number of nearest neighbor sinks to clip to
        clipping_buffer : float, optional
            Buffer to increase clipping radius by, by default 1.05
        radius : None | int, optional
            Force clipping radius if set to an int

        Returns
        -------
        radius : int
            Clipping radius in cost raster pixels
        x_feats : pd.DataFrame
            Substations, load centers, sinks, and nearest points on
            t-lines to SC point
        """
        logger.debug('Clipping features to sc_point {}'.format(sc_point.name))

        if len(self.sink_coords) > 2 or radius:
            row, col = sc_point[['row', 'col']].values

            if radius is None:
                _, pos = self.sink_tree.query([row, col], k=nn_sinks)
                radius = np.abs(self.sink_coords[pos] - np.array([row, col])
                                ).max()
                radius = int(np.ceil(radius * clipping_buffer))

            if radius:
                logger.debug('Using forced radius of {}'.format(radius))
            else:
                logger.debug('Radius to {} nearest sink is: {}'
                             .format(nn_sinks, radius))
            row_min = max(row - radius, 0)
            row_max = min(row + radius, self._shape[0])
            col_min = max(col - radius, 0)
            col_max = min(col + radius, self._shape[1])
            logger.debug('Extracting all transmission features in the row '
                         'slice {}:{} and column slice {}:{}'
                         .format(row_min, row_max, col_min, col_max))

            # Clip transmission features
            mask = self.features['row'] >= row_min
            mask &= self.features['row'] < row_max
            mask &= self.features['col'] >= col_min
            mask &= self.features['col'] < col_max
            sc_features = self.features.loc[mask].copy(deep=True)
            logger.debug('{} transmission features found in clipped area with '
                         'radius {}'
                         .format(len(sc_features), radius))
        else:
            sc_features = self.features.copy(deep=True)

        mask = self.features['max_volts'] >= tie_line_voltage
        sc_features = sc_features.loc[mask].copy(deep=True)
        logger.debug('{} transmission features found in clipped area with '
                     'minimum max voltage of {}'
                     .format(len(sc_features), tie_line_voltage))

        # Find t-lines connected to substations within clip
        logger.debug('Collecting transmission lines connected to substations')
        mask = sc_features['category'] == SUBSTATION_CAT
        if mask.any():
            trans_gids = sc_features.loc[mask, 'trans_gid'].values
            trans_gids = \
                np.concatenate(self.sub_lines_mapping.loc[trans_gids].values)
            trans_gids = np.unique(trans_gids)
            line_mask = self.features['trans_gid'].isin(trans_gids)
            trans_lines = self.features.loc[line_mask].copy(deep=True)
            line_mask = trans_lines['trans_gid'].isin(sc_features['trans_gid'])
            trans_lines = trans_lines.loc[~line_mask]
            logger.debug('Adding all {} transmission lines connected to '
                         'substations with minimum max voltage of {}'
                         .format(len(trans_lines), tie_line_voltage))
            sc_features = sc_features.append(trans_lines)

        return sc_features, radius

    def process_sc_points(self, capacity_class, sc_point_gids=None, nn_sinks=2,
                          clipping_buffer=1.05, barrier_mult=100,
                          max_workers=None, save_paths=False, radius=None,
                          mp_delay=3, simplify_geo=None):
        """
        Compute Least Cost Transmission for desired sc_points

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        sc_point_gids : list, optional
            List of sc_point_gids to connect to, by default connect to
            all
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to return least cost paths as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius if set to an int
        mp_delay : float, optional
            Delay in seconds between starting multi-process workers.
            Useful for reducing memory spike at working startup.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.GeoDataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        max_workers = os.cpu_count() if max_workers is None else max_workers

        if sc_point_gids is None:
            sc_point_gids = self.sc_points['sc_point_gid'].values

        tie_line_voltage = self._config.capacity_to_kv(capacity_class)

        if max_workers > 1:
            logger.info('Computing Least Cost Transmission for SC points in '
                        'parallel on {} workers'.format(max_workers))
            least_costs = self._process_multi_core(
                capacity_class,
                tie_line_voltage,
                sc_point_gids=sc_point_gids,
                nn_sinks=nn_sinks,
                clipping_buffer=clipping_buffer,
                barrier_mult=barrier_mult,
                save_paths=save_paths,
                radius=radius,
                mp_delay=mp_delay,
                simplify_geo=simplify_geo,
                max_workers=max_workers)
        else:
            logger.info('Computing Least Cost Transmission for {:,} SC points '
                        'in serial'.format(len(sc_point_gids)))
            least_costs = self._process_single_core(
                capacity_class,
                tie_line_voltage,
                sc_point_gids=sc_point_gids,
                nn_sinks=nn_sinks,
                clipping_buffer=clipping_buffer,
                barrier_mult=barrier_mult,
                save_paths=save_paths,
                radius=radius,
                simplify_geo=simplify_geo)

        least_costs = pd.concat(least_costs).sort_values(['sc_point_gid',
                                                          'trans_gid'])
        capacity_class = self._config._parse_cap_class(capacity_class)
        least_costs['max_cap'] = self._config['power_classes'][capacity_class]
        lcp_frac = (len(least_costs['sc_point_gid'].unique())
                    / len(sc_point_gids) * 100)
        logger.info('{:.4f}% of requested sc point gids were successfully '
                    'mapped to transmission features'.format(lcp_frac))

        return least_costs.reset_index(drop=True)

    def _process_multi_core(self, capacity_class, tie_line_voltage,
                            sc_point_gids, nn_sinks=2,
                            clipping_buffer=1.05, barrier_mult=100,
                            max_workers=2, save_paths=False, radius=None,
                            mp_delay=3, simplify_geo=None):
        """
        Compute Least Cost Transmission for desired sc_points using
        multiple cores.

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        tie_line_voltage : int
            Tie-line voltage (kV)
        sc_point_gids : list | set
            List of sc_point_gids to connect to, by default connect to
            all
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing
        save_paths : bool, optional
            Flag to return least cost paths as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius if set to an int
        mp_delay : float, optional
            Delay in seconds between starting multi-process workers.
            Useful for reducing memory spike at working startup.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.GeoDataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        least_costs = []
        num_jobs = 0
        loggers = [__name__, 'reV', 'reVX']
        with SpawnProcessPool(max_workers=max_workers, loggers=loggers) as exe:
            futures = []
            for _, sc_point in self.sc_points.iterrows():
                gid = sc_point['sc_point_gid']
                if gid in sc_point_gids:
                    sc_features, radius = self._clip_to_sc_point(
                        sc_point, tie_line_voltage, nn_sinks=nn_sinks,
                        clipping_buffer=clipping_buffer, radius=radius)
                    if sc_features.empty:
                        continue

                    future = exe.submit(TransCapCosts.run,
                                        self._cost_fpath,
                                        sc_point.copy(deep=True),
                                        sc_features, capacity_class,
                                        radius=radius,
                                        xmission_config=self._config,
                                        barrier_mult=barrier_mult,
                                        min_line_length=self._min_line_len,
                                        save_paths=save_paths,
                                        simplify_geo=simplify_geo)
                    futures.append(future)

                    num_jobs += 1
                    if num_jobs <= max_workers:
                        time.sleep(mp_delay)

            logger.debug('Completed kicking off {} jobs for {} workers.'
                         .format(num_jobs, max_workers))
            for i, future in enumerate(as_completed(futures), start=1):
                sc_costs = future.result()
                if sc_costs is not None:
                    least_costs.append(sc_costs)

                logger.info('SC point {} of {} complete!'
                            .format(i, len(futures)))
                log_mem(logger)

        return least_costs

    def _process_single_core(self, capacity_class, tie_line_voltage,
                             sc_point_gids, nn_sinks=2,
                             clipping_buffer=1.05, barrier_mult=100,
                             save_paths=False, radius=None,
                             simplify_geo=None):
        """
        Compute Least Cost Transmission for desired sc_points with a
        single core.

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        tie_line_voltage : int
            Tie-line voltage (kV)
        sc_point_gids : list | set
            List of sc_point_gids to connect to, by default connect to
            all
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        save_paths : bool, optional
            Flag to return least cost paths as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius if set to an int
        simplify_geo : float | None, optional
            If float, simplify geometries using this value

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.GeoDataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        least_costs = []
        for i, (_, sc_point) in enumerate(self.sc_points.iterrows(), start=1):
            gid = sc_point['sc_point_gid']
            if gid in sc_point_gids:
                sc_features, radius = self._clip_to_sc_point(
                    sc_point, tie_line_voltage, nn_sinks=nn_sinks,
                    clipping_buffer=clipping_buffer, radius=radius)
                if sc_features.empty:
                    continue

                sc_costs = TransCapCosts.run(
                    self._cost_fpath,
                    sc_point.copy(deep=True),
                    sc_features, capacity_class,
                    radius=radius,
                    xmission_config=self._config,
                    barrier_mult=barrier_mult,
                    min_line_length=self._min_line_len,
                    save_paths=save_paths,
                    simplify_geo=simplify_geo)

                if sc_costs is not None:
                    least_costs.append(sc_costs)

                logger.info('SC point {} of {} complete!'
                            .format(i, len(sc_point_gids)))
                log_mem(logger)
        return least_costs

    @classmethod
    def run(cls, cost_fpath, features_fpath, capacity_class, resolution=128,
            xmission_config=None, min_line_length=0, sc_point_gids=None,
            nn_sinks=2, clipping_buffer=1.05, barrier_mult=100,
            max_workers=None, save_paths=False, radius=None,
            simplify_geo=None):
        """
        Find Least Cost Transmission connections between desired
        sc_points to given transmission features for desired capacity
        class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        min_line_length : int | float, optional
            Minimum line length in km, by default 0.
        sc_point_gids : list, optional
            List of sc_point_gids to connect to, by default None
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to return least costs path as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius if set to an int
        simplify_geo : float | None, optional
            If float, simplify geometries using this value

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.DataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        ts = time.time()
        lcx = cls(cost_fpath, features_fpath, resolution=resolution,
                  xmission_config=xmission_config,
                  min_line_length=min_line_length)
        least_costs = lcx.process_sc_points(capacity_class,
                                            sc_point_gids=sc_point_gids,
                                            nn_sinks=nn_sinks,
                                            clipping_buffer=clipping_buffer,
                                            barrier_mult=barrier_mult,
                                            max_workers=max_workers,
                                            save_paths=save_paths,
                                            radius=radius,
                                            simplify_geo=simplify_geo)

        logger.info('{} connections were made to {} SC points in {:.4f} '
                    'minutes'
                    .format(len(least_costs),
                            len(least_costs['sc_point_gid'].unique()),
                            (time.time() - ts) / 60))

        return least_costs


class ReinforcedXmission(LeastCostXmission):
    """
    Compute Least Cost tie-line paths and full transmission cap cost
    for all supply curve points to all possible connections (substations
    within the SC balancing area).
    """

    def __init__(self, cost_fpath, features_fpath, balancing_areas_fpath,
                 resolution=128, xmission_config=None, min_line_length=0):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers.
        features_fpath : str
            Path to GeoPackage with transmission features. All features
            except substations will be dropped. This table must have a
            "ba_str" column which matches one of the "ba_str" ID's in
            the Balancing Areas GeoPackage.
        balancing_areas_fpath : str
            Path to GeoPackage with balancing areas. This table must
            have a "ba_str" column which matches the "ba_str" ID's in
            the Transmission Features GeoPackage.
        resolution : int, optional
            SC point resolution. By default, ``128``.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        min_line_length : int | float, optional
            Minimum line length in km. By default, ``0``.
        """
        super().__init__(cost_fpath=cost_fpath,
                         features_fpath=features_fpath,
                         resolution=resolution,
                         xmission_config=xmission_config,
                         min_line_length=min_line_length)
        self._ba = (gpd.read_file(balancing_areas_fpath)
                    .to_crs(self.features.crs))

    @staticmethod
    def _load_trans_feats(features_fpath):
        """Load existing substations from disk. """

        logger.debug('Loading substations...')
        substations = gpd.read_file(features_fpath)
        substations = substations[substations.category == SUBSTATION_CAT]
        substations = substations.reset_index(drop=True)
        substations = substations.drop(columns=['bgid', 'egid', 'cap_left'],
                                       errors='ignore')
        mapping = {'gid': 'trans_gid', 'trans_gids': 'trans_line_gids'}
        substations = substations.rename(columns=mapping)

        return substations, None

    def _clip_to_sc_point(self, sc_point, tie_line_voltage, nn_sinks=2,
                          clipping_buffer=1.05, radius=None):
        """Clip features to be substations in the BA of the sc point.  """
        logger.debug('Clipping features to sc_point {}'.format(sc_point.name))

        point = self.sc_points.loc[sc_point.name:sc_point.name].centroid
        ba_str = point.apply(ba_mapper(self._ba)).values[0]
        mask = self.features["ba_str"] == ba_str
        sc_features = self.features.loc[mask].copy(deep=True)

        mask = self.features['max_volts'] >= tie_line_voltage
        sc_features = sc_features.loc[mask].copy(deep=True)

        if sc_features.empty:
            return sc_features, None

        dists = (sc_features[['row', 'col']] - sc_point[['row', 'col']])
        radius = int(np.ceil(dists.abs().values.max() * clipping_buffer))
        logger.debug('{} transmission features found in clipped area of '
                     'radius {} with minimum max voltage of {}'
                     .format(len(sc_features), radius, tie_line_voltage))

        return sc_features, radius

    @classmethod
    def run(cls, cost_fpath, features_fpath, balancing_areas_fpath,
            capacity_class, resolution=128, xmission_config=None,
            min_line_length=0, sc_point_gids=None, clipping_buffer=1.05,
            barrier_mult=100, max_workers=None, save_paths=False,
            simplify_geo=None):
        """
        Find Least Cost Transmission connections between desired
        sc_points and substations in their balancing area.

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features. All features
            except substations will be dropped. This table must have a
            "ba_str" column which matches one of the "ba_str" ID's in
            the Balancing Areas GeoPackage.
        balancing_areas_fpath : str
            Path to GeoPackage with balancing areas. This table must
            have a "ba_str" column which matches the "ba_str" ID's in
            the Transmission Features GeoPackage.
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to.
        resolution : int, optional
            SC point resolution. By default, ``128``.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        min_line_length : int | float, optional
            Minimum line length in km. By default, ``0``.
        sc_point_gids : list, optional
            List of sc_point_gids to connect to. By default, ``None``,
            which processes all points.
        clipping_buffer : float, optional
            Buffer to expand clipping radius by. By default, ``1.05``.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        max_workers : int, optional
            Number of workers to use for processing. If 1 run in serial,
            if ``None`` use all available cores. By default, ``None``.
        save_paths : bool, optional
            Flag to save reinforcement line path as a multi-line
            geometry. By default, ``False``.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value.

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.DataFrame
            Least cost connections between all supply curve points and
            the substations in their balancing area with the given
            capacity class.
        """
        ts = time.time()
        lcx = cls(cost_fpath, features_fpath, balancing_areas_fpath,
                  resolution=resolution, xmission_config=xmission_config,
                  min_line_length=min_line_length)
        least_costs = lcx.process_sc_points(capacity_class,
                                            sc_point_gids=sc_point_gids,
                                            clipping_buffer=clipping_buffer,
                                            barrier_mult=barrier_mult,
                                            max_workers=max_workers,
                                            save_paths=save_paths,
                                            simplify_geo=simplify_geo)

        logger.info('{} connections were made to {} SC points in {:.4f} '
                    'minutes'
                    .format(len(least_costs),
                            len(least_costs['sc_point_gid'].unique()),
                            (time.time() - ts) / 60))

        return least_costs


def ba_mapper(ba):
    """Generate a function to map points to a BA.

    The returned mapping function maps a point to a unique "ba_str" from
    the input GeoPackage.

    Parameters
    ----------
    ba : gpd.GeoPackage
        GeoPackage defining the balancing areas. This table must have a
        "ba_str" column which and a geometry that defines each BA.
    """
    def _map_ba(point):
        """Find the balancing area ID for the input point. """
        return ba.loc[ba.distance(point).sort_values().index[0], "ba_str"]

    return _map_ba



# -*- coding: utf-8 -*-
"""
Module to compute least cost transmission paths and distances
"""
from concurrent.futures import as_completed
import geopandas as gpd
import logging
import numpy as np
import os
import pandas as pd
from pyproj.crs import CRS
import rasterio
import time
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

from reVX.least_cost_xmission.config import TRANS_LINE_CAT, SUBSTATION_CAT
from reVX.least_cost_xmission.trans_cap_costs import (TieLineCosts,
                                                      ReinforcementLineCosts)
from reVX.utilities import ExclusionsConverter

logger = logging.getLogger(__name__)


class LeastCostPaths:
    """
    Compute least cost paths between desired locations
    """
    REQUIRED_LAYERS = ['transmission_barrier']

    def __init__(self, cost_fpath, features_fpath, xmission_config=None):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        """
        self._check_layers(cost_fpath)
        self._config = TieLineCosts._parse_config(
            xmission_config=xmission_config)

        self._features, self._row_slice, self._col_slice, self._shape = \
            self._map_to_costs(cost_fpath, gpd.read_file(features_fpath))
        self._features = self._features.drop(columns='geometry')
        self._cost_fpath = cost_fpath
        self._start_feature_ind = 0

        logger.debug('{} initialized'.format(self))

    def __repr__(self):
        msg = ("{} to be computed for "
               .format(self.__class__.__name__))

        return msg

    @property
    def features(self):
        """
        Table of features to compute paths for

        Returns
        -------
        pandas.DataFrame
        """
        return self._features

    @classmethod
    def _check_layers(cls, cost_fpath):
        """
        Check to make sure the REQUIRED_LAYERS are in cost_fpath

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        """
        with ExclusionLayers(cost_fpath) as f:
            missing = []
            for lyr in cls.REQUIRED_LAYERS:
                if lyr not in f:
                    missing.append(lyr)

            if missing:
                msg = ("The following layers are required to compute Least "
                       "Cost Transmission but are missing from {}:\n{}"
                       .format(cost_fpath, missing))
                logger.error(msg)
                raise RuntimeError(msg)

    @staticmethod
    def _get_feature_cost_indices(features, crs, transform, shape):
        """
        Map features to cost row, col indices using rasterio transform

        Parameters
        ----------
        features : gpd.GeoDataFrame
            GeoDataFrame of features to map to cost raster
        crs : pyproj.crs.CRS
            CRS of cost raster
        transform : raster.Affine
            Transform of cost raster
        shape : tuple
            Cost raster shape

        Returns
        -------
        row : ndarray
            Vector of row indices for each feature
        col : ndarray
            Vector of col indices for each features
        mask : ndarray
            Boolean mask of features with indices outside of cost raster
        """
        feat_crs = features.crs.to_dict()
        cost_crs = crs.to_dict()
        bad_crs = ExclusionsConverter._check_crs(cost_crs, feat_crs)
        if bad_crs:
            msg = ('input crs ({}) does not match cost raster crs ({})'
                   ' and will be transformed!'.format(feat_crs, cost_crs))
            logger.warning(msg)
            warn(msg)
            features = features.to_crs(crs)

        logger.debug('Map features to cost raster')
        coords = features['geometry'].centroid
        row, col = rasterio.transform.rowcol(transform, coords.x.values,
                                             coords.y.values)
        row = np.array(row)
        col = np.array(col)

        # Remove features outside of the cost domain
        mask = row >= 0
        mask &= row < shape[0]
        mask &= col >= 0
        mask &= col < shape[1]

        return row, col, mask

    @staticmethod
    def _get_clip_slice(row, col, shape):
        """
        Clip cost raster to bounds of features

        Parameters
        ----------
        row : ndarray
            Vector of row indices
        col : ndarray
            Vector of col indices
        shape : tuple
            Full cost array shape

        Returns
        -------
        row_slice : slice
            Row slice to clip too
        col_slice : slice
            Col slice to clip too
        """
        row_slice = slice(max(row.min() - 1, 0), min(row.max() + 1, shape[0]))
        col_slice = slice(max(col.min() - 1, 0), min(col.max() + 1, shape[1]))

        return row_slice, col_slice

    @classmethod
    def _map_to_costs(cls, cost_fpath, features):
        """
        Map features to cost arrays

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features : gpd.GeoDataFrame
            GeoDataFrame of features to connect

        Returns
        -------
        features : gpd.GeoDataFrame
            Table of features to compute LeastCostPaths for
        row_slice : slice
            Clipping slice along axis-0 (rows)
        col_slice : slice
            Clipping slice along axis-1 (cols)
        shape : tuple
            Full cost raster shape
        """
        with ExclusionLayers(cost_fpath) as f:
            crs = CRS.from_string(f.crs)
            transform = rasterio.Affine(*f.profile['transform'])
            shape = f.shape

        row, col, mask = cls._get_feature_cost_indices(features, crs,
                                                       transform, shape)
        if any(~mask):
            msg = ("The following features are outside of the cost exclusion "
                   "domain and will be dropped:\n{}"
                   .format(features.loc[~mask]))
            logger.warning(msg)
            row = row[mask]
            col = col[mask]
            features = features.loc[mask].reset_index(drop=True)

        row_slice, col_slice = cls._get_clip_slice(row, col, shape)

        features['row'] = row - row_slice.start
        features['col'] = col - col_slice.start

        return features, row_slice, col_slice, shape

    @property
    def start_indices(self):
        """
        Tuple of (row_idx, col_idx) in the cost array indicating the
        start position of all paths to compute (typically, this is
        the centroid of the supply curve cell under consideration).
        Paths will be computed from this start location to each of the
        `end_indices`, which are also locations in the cost array
        (typically transmission feature locations).

        Returns
        -------
        tuple
        """
        start_indices = (self._features.loc[self._start_feature_ind,
                                            ['row', 'col']].values)
        if len(start_indices.shape) == 2:
            start_indices = start_indices[0]
        return start_indices

    @property
    def end_features(self):
        """
        GeoDataFrame containing the transmission features to compute the
        least cost paths to, starting from the `start_indices`
        (typically the centroid of the supply curve cell under
        consideration).

        Returns
        -------
        geopandas.GeoDataFrame
        """
        end_features = self._features.drop(index=self._start_feature_ind)
        end_features['start_index'] = self._start_feature_ind
        return end_features.reset_index(drop=False)

    @property
    def end_indices(self):
        """
        Tuple (row, col) index or list of (row, col) indices in the
        cost array indicating the end location(s) to compute least
        cost paths to (typically transmission feature locations).
        Paths are computed from the `start_indices` (typically the
        centroid of the supply curve cell under consideration) to each
        of the individual pairs of `end_indices`.

        Returns
        -------
        tuple | list
        """
        return self.end_features[['row', 'col']].values

    def process_least_cost_paths(self, capacity_class, barrier_mult=100,
                                 indices=None, max_workers=None,
                                 save_paths=False,):
        """
        Find Least Cost Paths between all pairs of provided features for
        the given tie-line capacity class

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each path or GeoDataFrame
            of length, cost, and geometry for each path
        """
        max_workers = os.cpu_count() if max_workers is None else max_workers
        indices = self.features.index if indices is None else indices
        least_cost_paths = []
        if max_workers > 1:
            logger.info('Computing Least Cost Paths in parallel on {} workers'
                        .format(max_workers))
            log_mem(logger)
            loggers = [__name__, 'reV', 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = {}
                for start in indices:
                    self._start_feature_ind = start
                    future = exe.submit(TieLineCosts.run, self._cost_fpath,
                                        self.start_indices, self.end_indices,
                                        capacity_class,
                                        self._row_slice, self._col_slice,
                                        barrier_mult=barrier_mult,
                                        save_paths=save_paths)
                    futures[future] = self.end_features

                for i, future in enumerate(as_completed(futures)):
                    end_features = futures[future]
                    end_features = end_features.drop(columns=['row', 'col'],
                                                     errors="ignore")
                    lcp = future.result()
                    lcp = pd.concat((end_features, lcp), axis=1)
                    least_cost_paths.append(lcp)
                    logger.debug('Least cost path {} of {} complete!'
                                 .format(i + 1, len(futures)))
                    log_mem(logger)
        else:
            logger.info('Computing Least Cost Paths in serial')
            log_mem(logger)
            i = 1
            for start in indices:
                self._start_feature_ind = start
                lcp = TieLineCosts.run(self._cost_fpath,
                                       self.start_indices, self.end_indices,
                                       capacity_class,
                                       self._row_slice, self._col_slice,
                                       barrier_mult=barrier_mult,
                                       save_paths=save_paths)
                end_features = self.end_features.drop(columns=['row', 'col'],
                                                      errors="ignore")
                lcp = pd.concat((end_features, lcp), axis=1)
                least_cost_paths.append(lcp)

                logger.debug('Least cost path {} of {} complete!'
                             .format(i, len(self.features)))
                log_mem(logger)
                i += 1

        least_cost_paths = pd.concat(least_cost_paths, ignore_index=True)

        return least_cost_paths

    @classmethod
    def run(cls, cost_fpath, features_fpath, capacity_class,
            xmission_config=None, barrier_mult=100, indices=None,
            max_workers=None, save_paths=False):
        """
        Find Least Cost Paths between all pairs of provided features for
        the given tie-line capacity class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each path or GeoDataFrame
            of length, cost, and geometry for each path
        """
        ts = time.time()
        lcp = cls(cost_fpath, features_fpath, xmission_config=xmission_config)
        least_cost_paths = lcp.process_least_cost_paths(
            capacity_class,
            barrier_mult=barrier_mult,
            indices=indices,
            save_paths=save_paths,
            max_workers=max_workers)

        logger.info('{} paths were computed in {:.4f} hours'
                    .format(len(least_cost_paths),
                            (time.time() - ts) / 3600))

        return least_cost_paths


class ReinforcementPaths(LeastCostPaths):
    """
    Compute reinforcement line paths between substations and a single
    balancing area network node.
    """
    def __init__(self, cost_fpath, features, transmission_lines,
                 xmission_config=None):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers.
        features : geopandas.GeoPackage
            GeoPackage with transmission features. The network
            node must be the first row of the GeoPackage - the rest
            should be substations that need to connect to that node.
        transmission_lines :dict
            Dictionary where the keys are the names of cost layers in
            the cost HDF5 file and values are arrays with the
            corresponding existing transmission lines rastered into
            them (i.e. array value is 1 at a pixel if there is a
            transmission line, otherwise 0). These arrays will be used
            to compute the reinforcement costs along existing
            transmission lines of differing voltages.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        """
        self._check_layers(cost_fpath)
        self._config = TieLineCosts._parse_config(
            xmission_config=xmission_config)

        self._features, self._row_slice, self._col_slice, self._shape = \
            self._map_to_costs(cost_fpath, features)
        self._features = self._features.drop(columns='geometry')
        self._cost_fpath = cost_fpath

        network_node = (self._features.iloc[0:1]
                        .dropna(axis="columns", how="all"))
        err_msg = "Network node was dropped during clipping!"
        assert "ba_str" in network_node, err_msg
        self._start_indices = network_node[['row', 'col']].values[0]
        self._features = (self._features.iloc[1:]
                          .reset_index(drop=True)
                          .dropna(axis="columns", how="all"))

        self._transmission_lines = transmission_lines

    @property
    def start_indices(self):
        """
        Tuple of (row_idx, col_idx) in the cost array indicating the
        start position of all reinforcement line paths to compute
        (typically, this is the location of the network node in the BA).
        Paths will be computed from this start location to each of the
        `end_indices`, which are also locations in the cost array
        (typically substations within the BA of the network node).

        Returns
        -------
        tuple
        """
        return self._start_indices

    @property
    def end_indices(self):
        """
        Tuple (row, col) index or list of (row, col) indices in the cost
        array indicating the end location(s) to compute reinforcement
        line paths to (typically substations within a single BA). Paths
        are computed from the `start_indices` (typically the network
        node of the BA) to each of the individual pairs of
        `end_indices`.

        Returns
        -------
        tuple | list
        """
        return self._features[['row', 'col']].values

    def process_least_cost_paths(self, capacity_class, barrier_mult=100,
                                 save_paths=False):
        """
        Find the reinforcement line paths between the network node and
        the substations for the given tie-line capacity class

        Parameters
        ----------
        capacity_class : str | int
            Transmission feature ``capacity_class`` to use for the
            'base' greenfield costs. 'Base' greenfield costs are only
            used if the reinforcement path *must* deviate from existing
            transmission lines. Typically, a capacity class of 400 MW
            (230kV transmission line) is used for the base greenfield
            costs.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        max_workers : int, optional
            Number of workers to use for processing. If 1 run in serial,
            if ``None`` use all available cores. By default, ``None``.
        save_paths : bool, optional
            Flag to save reinforcement line path as a multi-line
            geometry. By default, ``False``.

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each reinforcement line
            path or GeoDataFrame of length, cost, and geometry for each
            reinforcement line path.
        """

        logger.info('Computing reinforcement path costs for start index {}'
                    .format(self._start_indices))
        log_mem(logger)

        lcp = ReinforcementLineCosts.run(self._transmission_lines,
                                         self._cost_fpath,
                                         self.start_indices, self.end_indices,
                                         capacity_class,
                                         self._row_slice, self._col_slice,
                                         barrier_mult=barrier_mult,
                                         save_paths=save_paths)
        feats = self._features.drop(columns=['row', 'col'])
        least_cost_paths = pd.concat((feats, lcp), axis=1)

        return least_cost_paths.drop("index", axis="columns", errors="ignore")

    @classmethod
    def run(cls, cost_fpath, features_fpath, network_nodes_fpath,
            transmission_lines_fpath, capacity_class, xmission_config=None,
            barrier_mult=100, indices=None, save_paths=False):
        """
        Find the reinforcement line paths between the network node and
        the substations for the given tie-line capacity class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers.
        features_fpath : str
            Path to GeoPackage with transmission features. The network
            node must be the first row of the GeoPackage - the rest
            should be substations that need to connect to that node.
            This table must have a "ba_str" column which matches the
            "ba_str" ID of the network node to the "ba_str" of the
            substations that should connect to it.
        network_nodes_fpath : str
            Path to GeoPackage with network node endpoints. There should
            be exactly one endpoint for each balancing area, and each
            endpoint should have a "ba_str" column that identifies
            matches exactly one of the ID's in the balancing area
            GeoPackage to be used in downstream models.
        capacity_class : str | int
            Transmission feature ``capacity_class`` to use for the
            'base' greenfield costs. 'Base' greenfield costs are only
            used if the reinforcement path *must* deviate from existing
            transmission lines. Typically, a capacity class of 400 MW
            (230kV transmission line) is used for the base greenfield
            costs.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        max_workers : int, optional
            Number of workers to use for processing. If 1 run in serial,
            if ``None`` use all available cores. By default, ``None``.
        save_paths : bool, optional
            Flag to save reinforcement line path as a multi-line
            geometry. By default, ``False``.

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each reinforcement line
            path or GeoDataFrame of length, cost, and geometry for each
            reinforcement line path.
        """
        ts = time.time()
        least_cost_paths = []
        lcp_kwargs = {"capacity_class": capacity_class,
                      "barrier_mult": barrier_mult,
                      "save_paths": save_paths}
        with ExclusionLayers(cost_fpath) as f:
            cost_crs = CRS.from_string(f.crs)
            cost_shape = f.shape
            cost_transform = rasterio.Affine(*f.profile['transform'])

        features = gpd.read_file(features_fpath).to_crs(cost_crs)
        substations = (features[features.category == SUBSTATION_CAT]
                       .reset_index(drop=True)
                       .dropna(axis="columns", how="all"))

        lines = gpd.read_file(transmission_lines_fpath).to_crs(cost_crs)
        transmission_lines = (lines[lines.category == TRANS_LINE_CAT]
                              .reset_index(drop=True))

        transmission_lines = _rasterize_transmission(transmission_lines,
                                                     xmission_config,
                                                     cost_shape,
                                                     cost_transform)

        network_nodes = gpd.read_file(network_nodes_fpath).to_crs(cost_crs)
        indices = network_nodes.index if indices is None else indices
        for loop_ind, index in enumerate(indices, start=1):
            network_node = (network_nodes.iloc[index:index + 1]
                            .reset_index(drop=True))
            ba_str = network_node["ba_str"].values[0]
            node_substations = substations[substations["ba_str"] == ba_str]
            node_substations = node_substations.reset_index(drop=True)
            logger.debug('Working on {} substations in BA {}'
                         .format(len(node_substations), ba_str))
            node_features = pd.concat([network_node, node_substations])
            rp = cls(cost_fpath, node_features, transmission_lines,
                     xmission_config=xmission_config)
            node_least_cost_paths = rp.process_least_cost_paths(**lcp_kwargs)
            node_least_cost_paths['ba_str'] = ba_str
            least_cost_paths += [node_least_cost_paths]

            logger.debug('Computed {}/{} reinforcement paths'
                         .format(loop_ind, len(indices)))

        logger.info('{} paths were computed in {:.4f} hours'
                    .format(len(least_cost_paths), (time.time() - ts) / 3600))

        return pd.concat(least_cost_paths, ignore_index=True)


def _rasterize_transmission(transmission_lines, xmission_config, cost_shape,
                            cost_transform):
    """Rasterize transmission lines and assemble them into a dict. """

    transmission_lines_dict = {}
    capacities = sorted(xmission_config['power_classes'].values())
    v_min = 0
    for capacity in capacities[:-1]:
        v_max = xmission_config['power_to_voltage'][str(capacity)]
        curr_lines = transmission_lines[
            (transmission_lines["voltage"] > v_min)
            & (transmission_lines["voltage"] <= v_max)
        ]
        out = _rasterize_transmission_layer(curr_lines, cost_shape,
                                            cost_transform)
        transmission_lines_dict[int(capacity)] = out
        v_min = v_max

    curr_lines = transmission_lines[transmission_lines["voltage"] > v_min]
    out = _rasterize_transmission_layer(curr_lines, cost_shape, cost_transform)
    transmission_lines_dict[int(capacities[-1])] = out
    return transmission_lines_dict


def _rasterize_transmission_layer(transmission_lines, cost_shape,
                                  cost_transform):
    """Rasterize a single transmission layer. """
    shapes = [(geom, 1) for geom in transmission_lines["geometry"]
              if geom is not None]
    out = np.zeros(cost_shape, dtype='uint8')
    rasterio.features.rasterize(shapes=shapes, out=out, out_shape=out.shape,
                                fill=0, transform=cost_transform)

    return out
